/**
  *  \file client/dialogs/helpdlg.cc
  *  \brief Help Viewer
  *
  *  The PCC2 Help Viewer displays XML rendered into a RichDocument
  *  (as opposed to PCC 1.x, which displays preformatted text with
  *  custom control codes from a .res file).
  *
  *  This module provides user interface and control.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/helpdlg.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "io/streams.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/richdocument.h"
#include "ui/richhelp.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/xmlread.h"

namespace {
    enum {
        id_Back = 100
    };

    /** Help dialog. */
    class WHelpDialog : public UIWindow {
        ObjectHolder h;
        UIRichDocument doc;
        string_t currentPage;

        /* History: this is a ring buffer. historyCount contains the number
           of valid elements, historyHead points to the next usable (that is,
           unless the buffer is full, an unused) element. We limit history by
           simply not increasing historyCount above MAX. */
        struct History {
            string_t page;
            int top;
            RichDocument::linkid_t link;
        };
        enum { MAX = 32 };
        History history[MAX];
        int historyCount;
        int historyHead;

        void init();
        void setBackState();
        void pushHistory();
     public:
        WHelpDialog();
        void loadPage(string_t page);
        void onLinkClick(string_t page);
        void onBack();
    };
}

/** Constructor. */
WHelpDialog::WHelpDialog()
    : UIWindow("Help", UIWindow::GREEN_WINDOW, COLOR_GRAY),
      doc(font_sizes[FONT_NORMAL].scaledBy(40, 20), UIRichDocument::fl_Help),
      currentPage(),
      historyCount(0),
      historyHead(0)
{
    init();
}

/** Build the dialog. */
void
WHelpDialog::init()
{
    // VBox
    //   HBox
    //     UIRichDocument
    //     UIScrollbar
    //   HBox
    //     UIButton...
    setLayout(UIVBoxLayout::instance5);

    /* We use instance5 for g1 to have a little room between the document and the scrollbar.
       Normally, we don't have room between scrollee and scrollbar, but here it serves as
       a margin. */
    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(doc);
    g1.add(h.add(new UIScrollbar(doc)));
    doc.sig_link_click.add(this, &WHelpDialog::onLinkClick);

    g2.add(h.add(new UIButton(id_Back, _("Back"), SDLK_BACKSPACE)).addCall(this, &WHelpDialog::onBack));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());

    add(g1);
    add(g2);
    pack();

    setBackState();
}

/** Set state of "Back" button. It is only enabled when we actually
    have history. */
void
WHelpDialog::setBackState()
{
    setStateOf(id_Back, st_Disabled, historyCount == 0);
}

/** Push current location on the history. */
void
WHelpDialog::pushHistory()
{
    // Anything to save?
    if (currentPage.empty())
        return;

    // Fill in entry
    History& h = history[historyHead];
    h.page = currentPage;
    h.top  = doc.getTop();
    h.link = doc.getSelectedLink();

    // Advance
    historyHead = (historyHead+1) % MAX;
    if (historyCount < MAX)
        ++historyCount;
    setBackState();
}

/** Load a page. */
void
WHelpDialog::loadPage(string_t page)
{
    // Clear document
    RichDocument& d = doc.getDocument();
    d.clear();

    Ptr<Stream> file = ResManager::getInstance().getStream(ResId("pcc2help.xml"));
    if (!file) {
        /* Render error page */
        d.add(RichText(_("Error")).withAttribute(RichTextStyleAttribute::Big).withAttribute(RichTextStyleAttribute::Bold));
        d.addParagraph();
        d.add(_("The help file could not be opened."));
        d.finish();
    } else {
        /* Found. Construct everything needed to render it. */
        XmlReader rdr(*file);
        loadHelpPage(d, rdr, page);
    }

    // Update widget
    doc.handleDocumentUpdate();
    currentPage = page;
}

/** Handle click on a link. */
void
WHelpDialog::onLinkClick(string_t page)
{
    pushHistory();
    loadPage(page);
}

/** Handle "Back" button. */
void
WHelpDialog::onBack()
{
    if (historyCount > 0) {
        // Pop entry
        --historyCount;
        historyHead = (historyHead > 0 ? historyHead-1 : MAX-1);
        const History& h = history[historyHead];

        // Load that page
        loadPage(h.page);
        doc.scrollTo(h.top, 1);
        if (doc.getDocument().isLinkVisible(h.link, GfxRect(0, doc.getTopY(), doc.getExtent().w, doc.getExtent().h)))
            doc.setSelectedLink(h.link);
        setBackState();
    }
}

/** Help, main entry point. Displays the help viewer, starting on the
    specified page. */
void
doHelp(string_t page)
{
    WHelpDialog dlg;
    dlg.loadPage(page);

    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
}
