/**
  *  \file client/labels.cc
  *  \brief Object Label Expressions
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include <memory>
#include "client/labels.h"
#include "cpluslib/sigslot.h"
#include "game/config.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/pref.h"
#include "game/ship.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/expr.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/interror.h"
#include "int/parseexpr.h"
#include "int/simple.h"
#include "client/history.h"
#include "int/process.h"

/** Stores all object labels. */
static std::vector<string_t> labels[2];

/** Compiled label expressions. */
static IntBCORef label_exprs[2];

/** Configuration file options for label expressions. */
static ConfigStringOption label_cfg[2] = {
    ConfigStringOption(getUserPreferences(), "Label.Ship"),
    ConfigStringOption(getUserPreferences(), "Label.Planet")
};

static SignalConnection conn_turn;
static SignalConnection conn_edit;

/** Recompute planet labels.
    \param what what to recompute. ShipLabel, PlanetLabel, or -1 for both.
    \param everything true to recompute all labels, false to recompute only those whose object changed */
static void
recomputeLabels(int what, bool everything, bool trackErrors)
{
    string_t last_error;
    int num_evals = 0, num_errors = 0;

    /* Planets */
    if (what != ShipLabel) {
        for (int pid = 1; pid <= NUM_PLANETS; ++pid) {
            /* Check whether we have to recompute */
            GPlanet& p = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
            if (!everything && !p.isDirty())
                continue;

            /* Recompute */
            labels[PlanetLabel][pid-1].clear();
            if (p.isExistant() && label_exprs[PlanetLabel] != 0) {
                ++num_evals;
                IntExecutionContext exec("Temporary: Labels");
                exec.pushNewContext(new IntPlanetContext(pid));
                exec.pushNewFrame(new IntExecutionFrame(label_exprs[PlanetLabel]));
                if (runTemporaryProcess(exec)) {
                    labels[PlanetLabel][pid-1] = IntValue::toStringOf(exec.getResult(), false);
                } else if (exec.getState() == IntExecutionContext::psError) {
                    last_error = exec.getError().what();
                    ++num_errors;
                } else {
                    // ignore silently
                }
            }
        }
    }

    /* Ships */
    if (what != PlanetLabel) {
        for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
            /* Check whether we have to recompute */
            GShip& s = getDisplayedTurn().getCurrentUniverse().getShip(sid);
            if (!everything && !s.isDirty())
                continue;

            /* Recompute */
            labels[ShipLabel][sid-1].clear();
            if (s.isExistant() && label_exprs[ShipLabel] != 0) {
                ++num_evals;
                IntExecutionContext exec("Temporary: Labels");
                exec.pushNewContext(new IntShipContext(sid));
                exec.pushNewFrame(new IntExecutionFrame(label_exprs[ShipLabel]));
                if (runTemporaryProcess(exec)) {
                    labels[ShipLabel][sid-1] = IntValue::toStringOf(exec.getResult(), false);
                } else if (exec.getState() == IntExecutionContext::psError) {
                    last_error = exec.getError().what();
                    ++num_errors;
                } else {
                    // ignore silently
                }
            }
        }
    }

    /* If all our evaluations failed, throw an error */
    if (trackErrors && num_errors > 0 && num_errors == num_evals)
        throw IntError(last_error);
}

/** Universe change: called when an object (may) change in the current universe. */
static void
onUniverseChanged()
{
    recomputeLabels(-1, false, false);
}

/** Turn change: called when a new turn is displayed. Recompute everything and watch new turn. */
static void
onTurnChanged()
{
    recomputeLabels(-1, true, false);
    conn_edit = getDisplayedTurn().getCurrentUniverse().sig_pre_update.add(onUniverseChanged);
}

/** Initialize labels. Must be called after loading configuration,
    before displaying the map. */
static void
initLabels()
{
    labels[ShipLabel].clear();
    labels[ShipLabel].resize(NUM_SHIPS);
    labels[PlanetLabel].clear();
    labels[PlanetLabel].resize(NUM_PLANETS);

    label_exprs[ShipLabel]   = compileExpression(label_cfg[ShipLabel](), IntCompilationContext());
    label_exprs[PlanetLabel] = compileExpression(label_cfg[PlanetLabel](), IntCompilationContext());

    conn_turn = sig_turn_changed.add(onTurnChanged);
}

/** Get label.
    \param kind Kind of object to query
    \param id Object Id */
string_t
getLabel(LabelKind kind, int id)
{
    if (id > 0 && id <= int(labels[kind].size()))
        return labels[kind][id-1];
    else
        return string_t();
}

/** Get label expression.
    \param kind Kind of object to query */
string_t
getLabelExpr(LabelKind kind)
{
    return label_cfg[kind]();
}

/** Set label expression.
    \param kind Kind of object to query
    \param expr Expression
    \throw IntError if there is a problem with the expression */
void
setLabelExpr(LabelKind kind, string_t expr)
{
    IntTokenizer tok(expr);
    if (tok.getCurrentToken() == tok.tEnd) {
        /* empty expression */
        label_exprs[kind] = 0;
    } else {
        /* expression given, parse it */
        std::auto_ptr<IntExprNode> expr(IntExpressionParser(tok).parse());
        if (tok.getCurrentToken() != tok.tEnd)
            throw IntError::garbageAtEnd(true);

        IntBCORef bco = new IntBytecodeObject();
        expr->compileValue(*bco, IntCompilationContext());
        bco->relocate();
        label_exprs[kind] = bco;
    }

    label_cfg[kind].set(expr);
    recomputeLabels(kind, true, true);
}

// Constraints: after mod_pref, before setDisplayedTurn()
extern const ModuleInitializer mod_labels[] = {
    { OnEnterGame, initLabels },
    // FIXME { OnLeaveGame, doneLabels },
    { 0, 0 },
};
