/**
  *  \file main.cc
  *  \brief Client Main Module
  *
  *  This file contains the main function for the planets client.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#include <cstdlib>
#include "arch/platform.h"
#include "client/chart.h"
#include "client/chartusr.h"
#include "client/playerscr.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/load.h"
#include "game/mission.h"
#include "game/reg.h"
#include "game/save.h"
#include "game/spec.h"
#include "gfx/gfx.h"
#include "gfx/init.h"
#include "gfx/palette.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/cmdline.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/recode.h"
#include "version.h"
#include "ui/commands.h"
#include "util/confpars.h"
#include "game/db.h"
#include "game/stat.h"
#include "int/intglobal.h"
#include "game/utilparse.h"
#include "game/msg.h"
#include "game/team.h"
#include "game/playerset.h"
#include "int/if/initif.h"
#include "client/dialogs/searchdlg.h"
#include "client/labels.h"
#include "game/pref.h"
#include "ui/widgets/statictext.h"
#include "client/history.h"
#include "int/if/guiif.h"
#include "int/simple.h"
#include "game/gamedirs.h"
#include "game/sel.h"
#include "client/widgets/dirchoose.h"
#include "client/widgets/playerchoose.h"
#include "client/mainopts.h"
#include "client/screens/mkturnscreen.h"
#include "client/screens/unpackscreen.h"
#include "client/screens/sweepscreen.h"
#include "ui/widgets/alignwidget.h"
#include "client/widgets/imagepanel.h"
#include "resmgr/resmgr.h"
#include "game/storage.h"
#include "client/widgets/reginfo.h"
#include "game/msgglobal.h"
#include "ui/widgets/dircrumbs.h"
#include "ui/widgets/button.h"
#include "client/dialogs/imperial.h"
#include "client/scr-sim.h"
#include "ui/widgets/minimalbutton.h"
#include "client/dialogs/helpdlg.h"
#ifdef CONF_FLAK_SUPPORT
# include "flak/flakconf.h"
#endif

static const ModuleInitializer*const client_modules[] = {
    mod_base_gui,
    mod_int,
    mod_intgame_common,
    mod_intgame_gui,
    mod_pref,
    mod_reg,
    mod_config,
    mod_coord,
#ifdef CONF_FLAK_SUPPORT
    mod_flak,
#endif
    mod_racenames,
    mod_missions,
    mod_fcode,
    mod_spec,
    mod_chartusr,
    mod_searchdlg,
    mod_team,
    mod_history,
    mod_labels,
    mod_msg,
    mod_stat,
    mod_sel,
    mod_msgglobal,
    mod_imperial,
    mod_sim,
    0,
};

namespace {
    // FIXME: PCC2GameChooser and PCC2GameChooserWindow are almost
    // complete copies of WVcrGameChooser and WVcrGameChooserWindow.
    class PCC2GameChooser : public WPlayerChooser {
        GDirectoryOverview ov;
     public:
        PCC2GameChooser(int id);
        void scan(Ptr<Directory> dir);
        void clear()
            { ov.clear(); }
        int getDirectoryFlags() const
            { return ov.getGlobalFlags(); }
        const GDirectoryOverview& getOverview() const
            { return ov; }
    };

    class PCC2GameChooserWindow : public WImagePanel {
        WDirectoryChooser    dir_chooser;
        PCC2GameChooser      game_chooser;
        UIDirectoryCrumbTrail crumb_trail;
        WRegInfoDisplay      reg_info_display;
        ObjectHolder         holder;
        UIButton             btn_cd;
        UIButton             btn_open;
        UIButton             btn_unpack;
        UIButton             btn_maketurn;
        UIButton             btn_sweep;
        PCC2OptionParser&    parser;

        void init();
        void switchTo(UIBaseWidget& w);
        void updateButtons();
     public:
        PCC2GameChooserWindow(Ptr<Directory> dir, bool oneup, PCC2OptionParser& parser);
        ~PCC2GameChooserWindow();
        bool handleEvent(const UIEvent& e, bool second_pass);
        void onEnterGame();
        void onEnterDirectory();
        void onMaketurn();
        void onUnpack();
        void onSweep();
        void onExit();
        void onDirChange();
        void onDirMove();
        void onCrumbClick(int n);
    };
}

static bool
chooseGame(PCC2OptionParser& parser, Ptr<Directory> dir, bool oneup)
{
    PCC2GameChooserWindow win(dir, oneup, parser);
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);
    return rv == cm_Enter;
}

static bool
askUnpack()
{
    return UIMessageBox(_("There are new result files for this game. "
                          "Do you want to unpack them now?"),
                        _("PCC2")).doYesNoDialog();
}

/** Initialize.
    \param id      widget id */
PCC2GameChooser::PCC2GameChooser(int id)
    : WPlayerChooser(id)
{
    addCommand(_("[Change into this directory]"));
}

void
PCC2GameChooser::scan(Ptr<Directory> dir)
{
    ov.scan(*dir);
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (ov.getPlayerFlags(i) & GDirectoryOverview::have_Gamedir) {
            if (ov.getPlayerFlags(i) & GDirectoryOverview::have_NewRST)
                setInfo(i, true, _("new RST"));
            else
                setInfo(i, true, _("unpacked"));
        } else if (ov.getPlayerFlags(i) & GDirectoryOverview::have_RST) {
            if (ov.getPlayerFlags(i) & GDirectoryOverview::have_TRN)
                setInfo(i, true, _("RST + TRN"));
            else
                setInfo(i, true, _("RST"));
        }
    }
}

/** Initialize.
    \param dir     initial directory
    \param oneup   false: display directory as is;
                   true: display dir's parent, place cursor on dir */
PCC2GameChooserWindow::PCC2GameChooserWindow(Ptr<Directory> dir, bool oneup, PCC2OptionParser& parser)
    : WImagePanel(_("Choose Game"), requestPixmap(ResId("bg.choose", ui_root->getExtent().w))),
      dir_chooser(dir, oneup, 10),
      game_chooser(11),
      crumb_trail(),
      btn_cd(0, _("Change directory"), 0),
      btn_open(0, _("Open Game"), 0),
      btn_unpack(0, _("Unpack"), 'u'),
      btn_maketurn(0, _("Maketurn"), 'm'),
      btn_sweep(0, _("Sweep"), 's'),
      parser(parser)
{
    init();
}

void
PCC2GameChooserWindow::init()
{
    // VBox
    //   HBox h2
    //     VBox v
    //       UIStatic "Race"
    //       PCC2GameChooser
    //       UISpacer
    //     VBox v2
    //       UIStatic "Directory"
    //       WDirectoryChooser
    //       WRegInfoDisplay
    //       UISpacer
    //       HBox h3
    //         UISpacer
    //         UIButton "Unpack"
    //         UIButton "Maketurn"
    //         UIButton "Sweep"
    //   HBox h
    //     UIButton "cd"
    //     UISpacer
    //     UIButton "open"
    //     UIButton "cancel"
    setLayout(UIVBoxLayout::instance5);

    UIGroup& v      = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& v2     = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& h      = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& h2     = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& h3     = holder.add(new UIGroup(UIHBoxLayout::instance5));

    btn_cd.sig_fired.add(this, &PCC2GameChooserWindow::onEnterDirectory);

    btn_open.sig_fired.add(this, &PCC2GameChooserWindow::onEnterGame);
    game_chooser.sig_dblclick.add(this, &PCC2GameChooserWindow::onEnterGame);

    UIButton& btn_exit = holder.add(new UIButton(cm_Escape, _("Exit"), SDLK_ESCAPE));

    h.add(btn_cd);
    h.add(holder.add(new UISpacer()));
    h.add(btn_open);
    h.add(btn_exit);

    v.add(holder.add(new UIStaticText(_("Race:"), UIStaticText::StaticTitle)));
    v.add(game_chooser);
    v.add(reg_info_display);
    v.add(holder.add(new UISpacer()));

    v2.add(holder.add(new UIStaticText(_("Game directory:"), UIStaticText::StaticTitle)));
    v2.add(dir_chooser);

    btn_maketurn.sig_fired.add(this, &PCC2GameChooserWindow::onMaketurn);
    btn_unpack.sig_fired.add(this, &PCC2GameChooserWindow::onUnpack);
    btn_sweep.sig_fired.add(this, &PCC2GameChooserWindow::onSweep);
    btn_exit.sig_fired.add(this, &PCC2GameChooserWindow::onExit);

    h3.add(holder.add(new UISpacer()));
    h3.add(btn_unpack);
    h3.add(btn_maketurn);
    h3.add(btn_sweep);
    v.add(holder.add(new UISpacer()));
    v.add(h3);

    h2.add(v2);
    h2.add(v);

    UIGroup& g0 = holder.add(new UIGroup(UIVBoxLayout::instance5));
    g0.add(h2);
    g0.add(crumb_trail);
    g0.add(h);
    UIAlignedWidget& aw = holder.add(new UIAlignedWidget(g0, GfxPoint(1, 1), GfxPoint(10, 10)));
    add(aw);

    setExtent(ui_root->getExtent());

    /* Help button */
    {
        LayoutInfo li;
        UIMinimalButton& help = holder.add(new UIMinimalButton(cm_Help, _("Press \"H\" for help"), 'h'));
        help.getLayoutInfo(li);
        GfxRect pos(0, 0, li.pref_size.x, li.pref_size.y);
        pos.moveToEdge(ui_root->getExtent(), 2, 0, 5);
        help.setExtent(pos);
        help.addKey();
        add(help);
    }

    crumb_trail.setDirectory(dir_chooser.getCurrentDirectory());
    crumb_trail.sig_changed.add(this, &PCC2GameChooserWindow::onCrumbClick);
    dir_chooser.sig_dirchange.add(this, &PCC2GameChooserWindow::onDirChange);
    dir_chooser.sig_moved.add(this, &PCC2GameChooserWindow::onDirMove);

    onDirMove();
}

PCC2GameChooserWindow::~PCC2GameChooserWindow()
{ }

/** Switch to widget. Switches to widget by Id, if possible. */
void
PCC2GameChooserWindow::switchTo(UIBaseWidget& w)
{
    if (!w.hasState(st_Disabled))
        w.setFocusedWidget();
}

/** Update buttons. */
void
PCC2GameChooserWindow::updateButtons()
{
    const int dir_flags = game_chooser.getDirectoryFlags();

    // "Change directory": always enabled

    // "Open game": enabled when we have RST or game directory
    btn_open.setState(st_Disabled,
                      (dir_flags & (GDirectoryOverview::have_RST | GDirectoryOverview::have_Gamedir)) == 0);

    // "Unpack": enabled when we have some RSTs, old or new ones.
    btn_unpack.setState(st_Disabled,
                        (dir_flags & (GDirectoryOverview::have_RST | GDirectoryOverview::have_NewRST | GDirectoryOverview::have_OtherRST)) == 0);

    // "Sweep": enabled when we have a directory; sweep can get
    // rid of all remainders, even those we don't see in the browser.
    btn_sweep.setState(st_Disabled, dir_chooser.getSubdirectory() == 0);

    // "Maketurn": enabled when we have an unpacked directory
    btn_maketurn.setState(st_Disabled, (dir_flags & GDirectoryOverview::have_Gamedir) == 0);
}

/** Event handler. Handles Tab/Left/Right and Enter keys. */
bool
PCC2GameChooserWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
            if (dir_chooser.hasState(st_Focused))
                switchTo(game_chooser);
            else
                switchTo(dir_chooser);
            return true;
         case SDLK_RETURN:
            if (dir_chooser.hasState(st_Focused)) {
                if (game_chooser.isAnyRaceAccessible()) {
                    switchTo(game_chooser);
                } else {
                    onEnterDirectory();
                }
            } else {
                onEnterGame();
            }
            return true;
         case SDLK_RIGHT:
            if (game_chooser.hasState(st_Disabled)) {
                // There are no games here, which means we are on a directory We can enter
                onEnterDirectory();
            } else if (game_chooser.hasState(st_Focused) && game_chooser.getCurrent() >= NUM_PLAYERS) {
                // We are on the "[Enter directory]" entry
                onEnterDirectory();
            } else {
                // There are directories, so choose from those
                switchTo(game_chooser);
            }
            return true;
         case SDLK_LEFT:
            if (dir_chooser.hasState(st_Focused))
                dir_chooser.enterParent();
            else
                switchTo(dir_chooser);
            return true;
         case 'h':
         case SDLK_F1:
            doHelp("pcc2:gamesel");
            return true;
         case vk_Quit:
            stop(cm_Escape);
            putKey(vk_Quit);
            return true;
        }
    }
    return WImagePanel::handleEvent(e, second_pass);
}

void
PCC2GameChooserWindow::onEnterDirectory()
{
    dir_chooser.enterCurrent();
    switchTo(dir_chooser);
}

void
PCC2GameChooserWindow::onMaketurn()
{
    // FIXME: proper condition!
    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        initDirectories(d->getName(), Platform::getDataDirectory());
        performInitialisation(OnEnterDirectory, client_modules);
        doMaketurn();
        performInitialisation(OnLeftDirectory, client_modules);
        initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());
        performInitialisation(OnEnterNoGame, client_modules);
    }
}

void
PCC2GameChooserWindow::onUnpack()
{
    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        initDirectories(d->getName(), Platform::getDataDirectory());
        performInitialisation(OnEnterDirectory, client_modules);
        doUnpack(game_chooser.getOverview().getPlayersWhere(GDirectoryOverview::have_Gamedir));
        performInitialisation(OnLeftDirectory, client_modules);
        initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());
        performInitialisation(OnEnterNoGame, client_modules);
        onDirMove();
    }
}

void
PCC2GameChooserWindow::onSweep()
{
    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        initDirectories(d->getName(), Platform::getDataDirectory());
        performInitialisation(OnEnterDirectory, client_modules);
        doSweep(game_chooser.getOverview().getPlayersWhere(GDirectoryOverview::have_Conflict));
        performInitialisation(OnLeftDirectory, client_modules);
        initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());
        performInitialisation(OnEnterNoGame, client_modules);
        onDirMove();
    }
}

void
PCC2GameChooserWindow::onExit()
{
    if (UIMessageBox(_("Do you want to exit PCC2?"), _("PCC2")).doYesNoDialog())
        stop(1);
}

/** Handler for "Open game" button. If possible, set parameters as
    entered in the dialog and stop() it. */
void
PCC2GameChooserWindow::onEnterGame()
{
    // new RSTs?
    if (game_chooser.getOverview().getGlobalFlags() & GDirectoryOverview::have_NewRST) {
        if (askUnpack()) {
            onUnpack();
            return;
        }
    }

    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        if (game_chooser.getCurrent() >= 0 && game_chooser.getCurrent() < NUM_PLAYERS) {
            parser.setGameDirName(d->getName());
            parser.setPlayerId(game_chooser.getCurrent()+1);
            stop(cm_Enter);
        } else {
            onEnterDirectory();
        }
    }
}

void
PCC2GameChooserWindow::onDirChange()
{
    crumb_trail.setDirectory(dir_chooser.getCurrentDirectory());
    updateButtons();
}

void
PCC2GameChooserWindow::onDirMove()
{
    game_chooser.clear();
    game_chooser.setDirectory(dir_chooser.getSubdirectory());
    if (game_chooser.getOverview().getGlobalFlags() != 0) {
        /* if there are any flags set, this means we scanned a directory. */
        reg_info_display.initFromDirectory(*dir_chooser.getSubdirectory());
    } else {
        reg_info_display.clear();
    }
    updateButtons();
}

void
PCC2GameChooserWindow::onCrumbClick(int n)
{
    if (n != crumb_trail.getNumberOfItems()-1) {
        dir_chooser.enterDirectory(crumb_trail.getItem(n));
        dir_chooser.setFocusedWidget();
    }
}


static bool
loadGame(int player, int flags)
{
    // Construct storage instance
    if ((flags & GDirectoryOverview::have_Gamedir) != 0) {
        GGameDataStorage::instance = new GGameDirectoryStorage(player);
    } else {
        ASSERT((flags & GDirectoryOverview::have_RST) != 0);
        GGameDataStorage::instance = new GGameResultStorage(player, (flags & GDirectoryOverview::have_TRN) != 0);
    }

    // Load
    try {
        Ptr<Stream> pxy = spec_file_dir->openFile("xyplan.dat", Stream::C_READ);
        getCurrentTurn().init();
        getCurrentTurn().loadPlanetXY(*pxy);
        if (Ptr<Stream> cdb = game_file_dir->openFileNT(format("chart%d.cc", player), Stream::C_READ))
            loadChartDatabase(*cdb, getCurrentTurn());
        GGameDataStorage::instance->load(getCurrentTurn());
        getCurrentTurn().postprocess();
        return true;
    }
    catch (FileProblemException& fpe) {
        messageBox(format(_("Error loading turn: %s: %s")) << fpe.getFileName() << fpe.what(), _("Error"));
        return false;
    }
    catch (std::exception& e) {
        messageBox(format(_("Error loading turn: %s: %s")) << e.what(), _("Error"));
        return false;
    }
}

int main(int /*argc*/, char** argv)
{
    // Startup initialisation
    performInitialisation(OnStartup, client_modules);

    // Parse options
    PCC2OptionParser parser;
    Platform::initOptions(parser);
    initGraphicsOptions(parser);
    initUIOptions(parser);
    parser.parse(Platform::convertArgumentsToUtf8(argv));

    console.write(LOG_INFO, "[PCC2 v" PCC2_VERSION "]");
    try {
        initUIGlobals(0 /* FIXME: icon */, "Planets Command Center II (v" PCC2_VERSION ")");
        initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());

        /* Name must be absolute because we'll be playing tricks with it */
        parser.setGameDirName(Directory::getAbsoluteName(parser.getGameDirName()));

        /* The first directory given on the command line will be shown directly.
           Upon subsequent passes through the loop, the game directory will be the
           game just played, so we should view its parent in the game chooser,
           with the cursor on the game. In addition, "pcc-v2 dir/sub" will show
           "dir" looking at "sub" (i.e. select from the players in "sub"). */
        bool oneup = false;

        if (parser.getMode() == parser.mDefault) {
            /* Default mode: check directory */
            Ptr<Directory> dir = Directory::openDir(parser.getGameDirName());
            GDirectoryOverview ov;
            ov.scan(*dir);
            if ((ov.getGlobalFlags() & GDirectoryOverview::have_NewRST) != 0 && askUnpack()) {
                initDirectories(parser.getGameDirName(), Platform::getDataDirectory());
                performInitialisation(OnEnterDirectory, client_modules);
                doUnpack(ov.getPlayersWhere(GDirectoryOverview::have_Gamedir));
                performInitialisation(OnLeftDirectory, client_modules);
                initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());
                performInitialisation(OnEnterNoGame, client_modules);
                parser.setMode(parser.mChooseGame);
                oneup = true;
            } else if (int pid = ov.getDefaultPlayer()) {
                parser.setPlayerId(pid);
                parser.setMode(parser.mPlay);
            } else {
                parser.setMode(parser.mChooseGame);
                oneup = true;
            }
        }

        while (1) {
            /* Choose game? */
            if (parser.getMode() == parser.mChooseGame) {
                performInitialisation(OnEnterNoGame, client_modules);
                if (!chooseGame(parser, Directory::openDir(parser.getGameDirName()), oneup))
                    break;
                parser.setMode(parser.mPlay);
            }

            /* Play game? */
            oneup = true;
            initGameGlobals(parser.getGameDirName(), parser.getRootDirName());
            performInitialisation(OnEnterDirectory, client_modules);

            // Load standard library. Note that we should do that immediately after the first
            // OnEnterNoGame or OnEnterDirectory, but it doesn't matter as long as the script
            // interface as well as the game selector are incomplete.
            if (!executeFile(Directory::makeFileName(Platform::getResourceDirectory(), "core.q"), false))
                throw std::runtime_error(_("The script initialisation file could not be loaded, see above messages. "
                                           "PCC2 cannot continue."));
            executeFile(Directory::makeFileName(Platform::getUserSettingsDirectory(1), "pcc2init.q"), true);

            runHook("ENTERDIRECTORY");

            GDirectoryOverview ov;
            ov.scan(*game_file_dir);
            if (parser.getPlayerId() == 0) {
                /* skip */
            } else if (ov.getPlayerFlags(parser.getPlayerId()) == 0) {
                messageBox(format(_("No data for player %d available in this game directory."), parser.getPlayerId()),
                           _("Planets Command Center"));
            } else {
                console.write(LOG_INFO, _("=== Loading... ==="));
                setPlayerIds(parser.getPlayerId(), parser.getPlayerId());
                performInitialisation(OnEnterGame, client_modules);

                if (loadGame(parser.getPlayerId(), ov.getPlayerFlags(parser.getPlayerId()))) {
                    setDisplayedTurn(0);
                    performInitialisation(OnEnteredGame, client_modules);
                    runHook("LOAD");
                    if (getCurrentTurn().getDatabaseTurn() < getCurrentTurn().getTurnNumber())
                        runHook("NEWTURN");

                    // play
                    doPlayerScreen();

                    // saving the game
                    console.write(LOG_INFO, _("=== Saving... ==="));
                    performInitialisation(OnLeaveGame, client_modules);
                    GGameDataStorage::instance->save(getCurrentTurn());
                    if (Ptr<Stream> cdb = game_file_dir->openFileNT(format("chart%d.cc", parser.getPlayerId()),
                                                                    Stream::C_CREATE))
                        saveChartDatabase(*cdb, getCurrentTurn());
                    performInitialisation(OnLeftGame, client_modules);
                } else {
                    performInitialisation(OnLeftGame, client_modules);
                }
            }
            GGameDataStorage::instance = 0;
            performInitialisation(OnLeftDirectory, client_modules);
            parser.setMode(parser.mChooseGame);
        } /* while */

        performInitialisation(OnShutdown, client_modules);
        console.write(LOG_INFO, _("Good bye."));
    }
    catch (FileProblemException& e) {
        console << _("Error during initialisation: ") << e.getFileName()
                << ": " << e.what() << "\n";
        Platform::showOutput();
    }
    catch(std::exception& e) {
        console.write(LOG_ERROR, string_t(_("Error during initialisation: ")) + e.what());
        Platform::showOutput();
    }
    return 0;
}
