/**
  *  \file playvcr.cc
  *  \brief Standalone VCR Player
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <SDL.h>
#include "arch/platform.h"
#include "client/vcrplay.h"
#include "cpluslib/format.h"
#include "io/streams.h"
#include "game/config.h"
#include "game/spec.h"
#include "game/struct.h"
#include "game/vcrdb.h"
#include "gfx/init.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"
#include "sound/sound.h"
#include "game/pref.h"
#include "cpluslib/holder.h"
#include "defs.h"
#include "client/scr-sim.h"
#include "game/host.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/statictext.h"
#include "util/modinit.h"
#include "game/team.h"
#include "game/fcode.h"
#include "game/gamedirs.h"
#include "client/playvcropts.h"
#include "client/playvcrbg.h"
#include "version.h"
#include "client/widgets/imagepanel.h"
#include "ui/widgets/alignwidget.h"
#include "client/widgets/playerchoose.h"
#include "client/widgets/dirchoose.h"
#include "ui/widgets/dircrumbs.h"
#include "ui/widgets/button.h"

#ifdef CONF_FLAK_SUPPORT
#  include "flak/flakdefs.h"
#  include "flak/flakconf.h"
#  include "flak/flak.h"
#  include "flak/vis.h"
#  include "flak/dlg-select.h"
#endif

static const ModuleInitializer* const playvcr_modules[] = {
    mod_base_gui,
    mod_pref,
    mod_racenames,
    mod_spec,
    mod_team,
    mod_fcode,
    mod_config,
#ifdef CONF_FLAK_SUPPORT
    mod_flak,
#endif
    0
};

/** Initialize. This is the VCR's equivalent to initGameGlobals. It
    initializes the directories and forgets the learned host type. 
    Unlike initGameGlobals, it doesn't set up a turn. */
static void
initVcrGlobals(string_t gamedir, string_t rootdir)
{
    initDirectories(gamedir, rootdir);
    host.clear();
}

/*************************** Directory Chooser ***************************/

namespace {
    /** Game chooser for VCR. Displays the VCR count for each race. */
    class WVcrGameChooser : public WPlayerChooser {
        string_t file_name[NUM_PLAYERS];
     public:
        WVcrGameChooser(int id);
        void scan(Ptr<Directory> dir);
        string_t getFileName(int player) const
            { return file_name[player-1]; }
    };

    /** Game chooser window. */
    class WVcrGameChooserWindow : public WVcrBackgroundWindow {
        WDirectoryChooser    dir_chooser;
        WVcrGameChooser      game_chooser;
        UIDirectoryCrumbTrail crumb_trail;
        ObjectHolder         holder;
        UIButton             btn_cd;
        UIButton             btn_open;
        PlayvcrOptionParser& parser;

        void init();
        void switchTo(UIBaseWidget& w);
        void updateButtons();
     public:
        WVcrGameChooserWindow(Ptr<Directory> dir, bool oneup, PlayvcrOptionParser& parser);
        ~WVcrGameChooserWindow();
        bool handleEvent(const UIEvent& e, bool second_pass);
        void onEnterDirectory();
        void onEnterGame();
        void onDirChange();
        void onDirMove();
        void onCrumbClick(int n);
        void onSimulate();
    };
}

static void doStandaloneSimulator();
static void initVcrGlobals(string_t gamedir, string_t rootdir);

/** Initialize.
    \param id      widget id */
WVcrGameChooser::WVcrGameChooser(int id)
    : WPlayerChooser(id)
{
    addCommand(_("[Change into this directory]"));
}

void
WVcrGameChooser::scan(Ptr<Directory> dir)
{
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        Ptr<Stream> s;

#ifdef CONF_FLAK_SUPPORT
        using namespace std;
        file_name[i-1] = format("flak%d.dat", i);
        s = dir->openFileNT(file_name[i-1], Stream::C_READ);
        TFlakHeader header;
        if (s && getStructure(*s, header) && memcmp(header.magic, FLAK_MAGIC, sizeof(header.magic)) == 0
            && header.num_battles > 0)
        {
            setInfo(i, true, format(_("%d fight%!1{s%}"), header.num_battles));
            continue;
        }
#endif

        file_name[i-1] = format("vcr%d.dat", i);
        s = dir->openFileNT(file_name[i-1], Stream::C_READ);
        TInt16 value;
        if (s && getStructure(*s, value) && value.value >= 0) {
            setInfo(i, true, format(_("%d fight%!1{s%}"), value.value));
        }
    }
}

/** Initialize.
    \param dir     initial directory
    \param oneup   false: display directory as is;
                   true: display dir's parent, place cursor on dir */
WVcrGameChooserWindow::WVcrGameChooserWindow(Ptr<Directory> dir, bool oneup, PlayvcrOptionParser& parser)
    : WVcrBackgroundWindow(),
      dir_chooser(dir, oneup, 10),
      game_chooser(11),
      btn_cd(0, _("Change directory"), 0),
      btn_open(0, _("View VCRs"), 0),
      parser(parser)
{
    init();
}

WVcrGameChooserWindow::~WVcrGameChooserWindow()
{ }

void
WVcrGameChooserWindow::init()
{
    // VBox g0
    //   HBox h2
    //     VBox v
    //       UIStatic "Directory"
    //       WDirectoryChooser
    //     VBox v2
    //       UIStatic "Race"
    //       WGamedirChooser
    //       UISpacer
    //   UIDirectoryCrumbTrail
    //   HBox h
    //     UIButton "cd"
    //     UISpacer
    //     UIButton "open"
    //     UIButton "cancel"

    /* Build the dialog */
    UIGroup& g0 = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& v  = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& v2 = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& h  = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& h2 = holder.add(new UIGroup(UIHBoxLayout::instance5));

    v.add(holder.add(new UIStaticText(_("Game directory:"), UIStaticText::StaticTitle)));
    v.add(dir_chooser);

    v2.add(holder.add(new UIStaticText(_("Race:"), UIStaticText::StaticTitle)));
    v2.add(game_chooser);
    v2.add(holder.add(new UISpacer()));

    h2.add(v);
    h2.add(v2);

    h.add(btn_cd);
    h.add(holder.add(new UISpacer()));
    h.add(btn_open);
    h.add(holder.add(new UIButton(cm_Escape, _("Exit"), SDLK_ESCAPE)).addStop());

    g0.add(h2);
    g0.add(crumb_trail);
    g0.add(h);

    add(holder.add(new UIAlignedWidget(g0, GfxPoint(1, 1), GfxPoint(10, 10))));
    setExtent(ui_root->getExtent());

    /* Connect events */
    btn_cd.sig_fired.add(this, &WVcrGameChooserWindow::onEnterDirectory);
    btn_open.sig_fired.add(this, &WVcrGameChooserWindow::onEnterGame);
    dir_chooser.sig_dirchange.add(this, &WVcrGameChooserWindow::onDirChange);
    dir_chooser.sig_moved.add(this, &WVcrGameChooserWindow::onDirMove);
    crumb_trail.sig_changed.add(this, &WVcrGameChooserWindow::onCrumbClick);
    game_chooser.sig_dblclick.add(this, &WVcrGameChooserWindow::onEnterGame);

    /* Synthetic movement action to update the player chooser */
    onDirChange();
    onDirMove();

    dir_chooser.setFocusedWidget();

    /* If the parameter set contains a file we understand, move cursor there */
    /* Caveat emptor: if something happens to resize the whole thing afterwards,
       the listboxes will emit sig_changed, which will cause onDirMove to be
       called, which resets the game_chooser's position. */
    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            if (Directory::makeFileName(d->getName(), game_chooser.getFileName(i)) == parser.getFileName()) {
                game_chooser.scrollTo(i-1, 1);
                switchTo(game_chooser);
                break;
            }
        }
    }
}

void
WVcrGameChooserWindow::switchTo(UIBaseWidget& w)
{
    if (!w.hasState(st_Disabled))
        w.setFocusedWidget();
}

void
WVcrGameChooserWindow::updateButtons()
{
    btn_open.setState(st_Disabled, game_chooser.hasState(st_Disabled));
}

void
WVcrGameChooserWindow::onEnterDirectory()
{
    dir_chooser.enterCurrent();
    switchTo(dir_chooser);
}

void
WVcrGameChooserWindow::onEnterGame()
{
    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        if (game_chooser.getCurrent() >= 0 && game_chooser.getCurrent() < NUM_PLAYERS) {
            // User has chosen a race. Set up the parameter set and close the dialog.
            parser.setDirName(d->getName());
            parser.setFileName(Directory::makeFileName(d->getName(), game_chooser.getFileName(game_chooser.getCurrent()+1)));
            stop(cm_Enter);
        } else {
            // User used the "Enter directory" item.
            onEnterDirectory();
        }
    }
}

void
WVcrGameChooserWindow::onDirChange()
{
    crumb_trail.setDirectory(dir_chooser.getCurrentDirectory());
    updateButtons();
}

void
WVcrGameChooserWindow::onDirMove()
{
    game_chooser.setDirectory(dir_chooser.getSubdirectory());
    updateButtons();
}

void
WVcrGameChooserWindow::onCrumbClick(int n)
{
    if (n != crumb_trail.getNumberOfItems()-1) {
        dir_chooser.enterDirectory(crumb_trail.getItem(n));
        dir_chooser.setFocusedWidget();
    }
}

/** Standalone simulator invocation from the Game Chooser Window. This
    assumes the state machine to be in the "NoGame" state, and brings
    it back there.*/
void
WVcrGameChooserWindow::onSimulate()
{
    // Figure out current directory
    Ptr<Directory> d = dir_chooser.getSubdirectory();
    if (!d || !d->getName().size())
        d = dir_chooser.getCurrentDirectory();

    // If we have a directory, play
    if (d && d->getName().size()) {
        initVcrGlobals(d->getName(), Platform::getDataDirectory());
        performInitialisation(OnEnterDirectory | OnEnterGame, playvcr_modules);

        doStandaloneSimulator();

        initVcrGlobals(Platform::getDataDirectory(), Platform::getDataDirectory());
        performInitialisation(OnLeftGame | OnLeftDirectory | OnEnterNoGame, playvcr_modules);
    }
}

/** Event handler. Handles Tab/Left/Right and Enter keys. */
bool
WVcrGameChooserWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_TAB:
         case SDLK_TAB + ss_Shift:
            if (dir_chooser.hasState(st_Focused))
                switchTo(game_chooser);
            else
                switchTo(dir_chooser);
            return true;
         case SDLK_RETURN:
            if (dir_chooser.hasState(st_Focused)) {
                if (game_chooser.isAnyRaceAccessible()) {
                    switchTo(game_chooser);
                } else {
                    onEnterDirectory();
                }
            } else {
                onEnterGame();
            }
            return true;
         case SDLK_RIGHT:
            if (game_chooser.hasState(st_Disabled))
                dir_chooser.enterCurrent();
            else
                switchTo(game_chooser);
            return true;
         case SDLK_LEFT:
            if (dir_chooser.hasState(st_Focused))
                dir_chooser.enterParent();
            else
                switchTo(dir_chooser);
            return true;
         case vk_Quit:
            stop(cm_Escape);
            putKey(vk_Quit);
            return true;
        }
    }
    return WVcrBackgroundWindow::handleEvent(e, second_pass);
}

static const UIKeyTranslator::Item VCR_SEL_KEYS[] = {
    { SDLK_UP,      cm_VCR_Prev, false },
    { SDLK_DOWN,    cm_VCR_Next, false },
    { vk_WheelDown, cm_VCR_Next, false },
    { vk_WheelUp,   cm_VCR_Prev, false },
    { vk_Quit,      cm_Escape,   false }
};

/** Play VCRs.
    \returns true if user chose a different file; main code should reload
    data and restart playing. */
static bool
playVcrs(GClassicVcrDatabase& db, int index)
{
    ObjectHolder holder;
    WVcrSelector sel(db);
    WVcrSelectorWindow w(sel);
    UIButton f(1, _("F - File"), 'f');
    f.addStop();

    if (index)
        sel.selectVcr(index-1);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(holder.add(new UIButton(cm_VCR_Prev, "\xE2\x86\x91", '-')).addCommand());
    but.add(holder.add(new UIButton(cm_VCR_Next, "\xE2\x86\x93", '+')).addCommand());
    but.add(holder.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)).addCommand());
    but.add(holder.add(new UISpacer()));
    but.add(f);
    but.add(holder.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE)).addStop());

    w.setLayout(UIVBoxLayout::instance5);
    w.setState(UIWindow::st_Modal, false);
    w.add(sel);
    w.add(but);
    w.pack();
    ui_root->centerWidget(w);
    ui_root->add(w);
    if (index)
        sel.playIt();

    bool rv = (ui_root->run() != cm_Escape);
    ui_root->close(w);

    return rv;
}

#ifdef CONF_FLAK_SUPPORT
namespace {
    class FlakVcrSelectorWindow : public UIWindow {
        FlakVcrSelector& sel;
     public:
        FlakVcrSelectorWindow(FlakVcrSelector& sel)
            : UIWindow(_("Visual Combat Recorder"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
              sel(sel)
            { }
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void playCurrent();
    };
}

bool
FlakVcrSelectorWindow::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case cm_VCR_Next:
        sel.next(cmd.info);
        return true;
     case cm_VCR_Prev:
        sel.previous(cmd.info);
        return true;
     case cm_Enter:
        playCurrent();
        return true;
     case cm_VCR_LeftInfo:
        sel.showInfo();
        return true;
    }
    return UIWindow::handleCommand(cmd, sender);
}

void
FlakVcrSelectorWindow::playCurrent()
{
    playFlakBattle(sel.getDatabase().getBattle(sel.getCurrent()));
}

static bool
playFlak(GFlakVcrDatabase& db, int index)
{
    ObjectHolder h;

    FlakVcrSelector sel(db);

    FlakVcrSelectorWindow w(sel);
    UIButton f(1, _("F - File"), 'f');
    f.addStop();

    UIGroup but(UIHBoxLayout::instance5);
    but.add(h.add(new UIButton(cm_VCR_Prev, "\xE2\x86\x91", '-')).addCommand());
    but.add(h.add(new UIButton(cm_VCR_Next, "\xE2\x86\x93", '+')).addCommand());
    but.add(h.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)).addCommand());
    but.add(h.add(new UISpacer()));
    but.add(f);
    but.add(h.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE)).addStop());

    UIGroup sel1(UIHBoxLayout::instance5), sel2(UIVBoxLayout::instance5);

    if (index)
        sel.setCurrent(index - 1);
    sel1.add(sel);
    sel1.add(sel2);
    sel2.add(h.add(new UIButton(cm_VCR_LeftInfo, "L", 'l')).addCommand());
    sel2.add(h.add(new UISpacer()));

    w.setLayout(UIVBoxLayout::instance5);
    w.setState(UIWindow::st_Modal, false);
    w.add(sel1);
    w.add(but);
    w.add(h.add(new UIKeyTranslator(VCR_SEL_KEYS, countof(VCR_SEL_KEYS))));
    w.pack();

    ui_root->centerWidget(w);
    ui_root->add(w);
    if (index)
        w.playCurrent();
    bool rv = (ui_root->run() != cm_Escape);
    ui_root->close(w);

    return rv;
}
#endif

static void
doStandaloneSimulator()
{
    GSimulatorNullInterface null_if;
    WSimScreen::sim_opts.loadDefaults(host.isPHost() ? GSimOptions::VcrPHost4 : GSimOptions::VcrHost);
    doSimulator(null_if);
}

static bool
doChooseDirectory(PlayvcrOptionParser& parser, bool oneup)
{
    Ptr<Directory> dir = Directory::openDir(parser.getDirName());
    WVcrGameChooserWindow w(dir, oneup, parser);
    return ui_root->runModal(w) == cm_Enter;
}

class PlayVcrVisitor : public GVcrDatabaseVisitor {
    int index;
    bool rv;
 public:
    PlayVcrVisitor(int index)
        : index(index) { }
    void visitClassicDatabase(GClassicVcrDatabase& db)
        {
            rv = playVcrs(db, index);
        }
    void visitFlakDatabase(GFlakVcrDatabase& db)
        {
#ifdef CONF_FLAK_SUPPORT
            rv = playFlak(db, index);
#endif
        }
    bool doContinue() const
        { return rv; }
};

/** Play a file.
    \param parser PlayvcrOptionParser containing file and directory name
    \retval true to stay in the file chooser
    \retval false to exit */
static bool
doPlayFile(PlayvcrOptionParser& parser)
{
    Ptr<Stream> str = Stream::openFile(parser.getFileName(), Stream::C_READ);

    /* Try interpreting the file as a saved simulation setup */
    try {
        GSimState tmp;
        tmp.load(*str);
        WSimScreen::sim_state.swap(tmp);
        doStandaloneSimulator();
        return false;
    }
    catch (FileFormatException& e) { }

    /* Open background window */
    class BackgroundPanel : public WVcrBackgroundWindow {
     public:
        void onSimulate()
            {
                doStandaloneSimulator();
            }
    };
    BackgroundPanel background;
    background.setExtent(ui_root->getExtent());
    ui_root->add(background);

    /* Try to load the file as a VCR file */
    Ptr<GVcrDatabase> db = 0;
    try {
#ifdef CONF_FLAK_SUPPORT
        TFlakHeader hdr;
        str->seek(0);
        if (getStructure(*str, hdr) && std::memcmp(hdr.magic, FLAK_MAGIC, sizeof(hdr.magic)) == 0) {
            str->seek(0);
            db = new GFlakVcrDatabase(*str);
        }
#endif
        if (!db) {
            str->seek(0);
            db = new GClassicVcrDatabase(*str);
        }
    }
    catch (FileFormatException& e) {
        messageBox(format(_("Unable to load file `%s': %s"), e.getFileName(), e.what()),
                   _("VCR Player"));
        return true;
    }
    str = 0;

    /* Play it */
    console.write(LOG_INFO, format(_("Loaded %d record%!1{s%} from `%s'"), db->getCount(), parser.getFileName()));
    if (db->getCount() == 0) {
        /* No battles to play */
        messageBox(format(_("The file `%s' does not contain any battles."),
                          parser.getFileName()),
                   _("VCR Player"));
        return true;
    } else {
        /* Check whether we were given a record number */
        int index = 0;
        if (parser.getBattleNumber() > 0) {
            if (parser.getBattleNumber() > db->getCount()) {
                messageBox(format(_("There are only %d battles in this VCR file."), db->getCount()),
                           _("VCR Player"));
            } else {
                index = parser.getBattleNumber();
            }
        }

        PlayVcrVisitor vis(index);
        db->visit(vis);
        return vis.doContinue();
    }
}

int
main(int, char** argv)
{
    // Startup
    performInitialisation(OnStartup, playvcr_modules);

    // Options
    PlayvcrOptionParser parser;
    Platform::initOptions(parser);
    initGraphicsOptions(parser);
    initUIOptions(parser);
    parser.parse(Platform::convertArgumentsToUtf8(argv));

    console.write(LOG_INFO, "[PlayVCR v" PCC2_VERSION "]");
    try {
        initUIGlobals("playvcr.bmp", _("PCC2 VCR Player"));
        loadSoundConfig();          // this will also initialize sound

        /* Names must be absolute because we'll be playing tricks */
        parser.setDirName(Directory::getAbsoluteName(parser.getDirName()));
        parser.setFileName(Directory::getAbsoluteName(parser.getFileName()));
        bool first = true;

        while (1) {
            if (parser.getMode() == parser.mChooseFile) {
                // Choose directory
                initVcrGlobals(Platform::getDataDirectory(), Platform::getDataDirectory());
                performInitialisation(OnEnterNoGame, playvcr_modules);
                if (doChooseDirectory(parser, !first)) {
                    // Continue
                    parser.setMode(parser.mPlayFile);
                } else {
                    // Exit
                    break;
                }
            }
            if (parser.getMode() == parser.mPlayFile) {
                // Play file (VCR or simulator)
                initVcrGlobals(parser.getDirName(), Platform::getDataDirectory());
                performInitialisation(OnEnterDirectory | OnEnterGame, playvcr_modules);

                bool result = doPlayFile(parser);

                performInitialisation(OnLeftGame | OnLeftDirectory, playvcr_modules);
                if (result) {
                    // Let's continue
                    parser.setMode(parser.mChooseFile);
                    parser.setBattleNumber(0);
                } else {
                    // Exit
                    break;
                }
            }
            if (parser.getMode() == parser.mSimulate) {
                // Simulator in a directory
                initVcrGlobals(parser.getDirName(), Platform::getDataDirectory());
                performInitialisation(OnEnterDirectory | OnEnterGame, playvcr_modules);
                doStandaloneSimulator();
                performInitialisation(OnLeftGame | OnLeftDirectory, playvcr_modules);
                // This is the command-line simulator; exit.
                break;
            }
            first = false;
        }
        performInitialisation(OnShutdown, playvcr_modules);
    }
    catch (FileProblemException& e) {
        console << _("Error during initialisation: ") << e.getFileName()
                << ": " << e.what() << "\n";
        Platform::showOutput();
    }
    catch (std::exception& e) {
        console << _("Error during initialisation: ") << e.what() << "\n";
        Platform::showOutput();
    }
    return 0;
}
