/**
  *  \file client/widgets/combatdiagram.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/widgets/combatdiagram.h"
#include "game/vcrdb.h"
#include "game/vcr.h"
#include "util/progress.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "client/widgets/racelist.h"
#include "client/marks.h"
#include "util/nls.h"
#include "io/io.h"
#include "util/recode.h"
#include "cpluslib/format.h"

static const int TEXT_EMS = 20;
static const int RIGHT_PX = 10;

static bool
SortGroups(const WCombatDiagram::Item& a, const WCombatDiagram::Item& b)
{
    if (a.group_id == b.group_id)
        return a.sequence < b.sequence;
    else
        return a.group_id < b.group_id;
}

/***************************** WCombatDiagram ****************************/

WCombatDiagram::WCombatDiagram()
{ }

void
WCombatDiagram::init(GClassicVcrDatabase& db, ProgressMonitor& prog)
{
    /* Clear old data */
    lines.clear();

    /* We may need a VCR player, so set one up */
    Ptr<VcrPlayer> player = db.createPlayer();
    if (player)
        player->setVisualizer(new VcrNullVisualizer());

    /* Check all VCRs */
    int group_counter = 0;
    int sequence      = 0;
    for (int i = 0; i < db.getCount(); ++i) {
        prog.tick();
        GVcrEntry& entry = db.getEntry(i);

        /* Check whether we must compute this entry's result first */
        if (entry.getStatus() == 0) {
            TVcr vcr_tmp = entry.getVcr();
            if (!player || !player->setCapabilities(db.getCapabilities()) || player->checkVcr(vcr_tmp)) {
                /* Cannot be played */
            } else {
                /* Play it */
                player->playVcr(vcr_tmp);
                entry.setResultFromPlayer(*player);
            }
        }

        /* Assimilate */
        std::vector<WCombatDiagram::Item>::const_iterator
            lptr = findObject(entry.isPlanet(false), entry.getVcr().objects[0].id),
            rptr = findObject(entry.isPlanet(true),  entry.getVcr().objects[1].id);
        if (lptr != lines.end() && rptr != lines.end()) {
            /* Both items are already known. Merge the groups if necessary. */
            if (lptr->group_id < rptr->group_id) {
                renameGroup(rptr->group_id, lptr->group_id);
            } else {
                renameGroup(lptr->group_id, rptr->group_id);
            }
        } else if (lptr != lines.end()) {
            /* Left is known, right is not */
            lines.push_back(Item(entry.isPlanet(true),  entry.getVcr().objects[1].id, lptr->group_id, ++sequence, i));
        } else if (rptr != lines.end()) {
            /* Right is known, left is not */
            lines.push_back(Item(entry.isPlanet(false), entry.getVcr().objects[0].id, rptr->group_id, ++sequence, i));
        } else {
            /* Neither is known */
            int group_id = ++group_counter;
            lines.push_back(Item(entry.isPlanet(true),  entry.getVcr().objects[1].id, group_id, ++sequence, i));
            lines.push_back(Item(entry.isPlanet(false), entry.getVcr().objects[0].id, group_id, ++sequence, i));
        }
    }

    /* Sort */
    std::sort(lines.begin(), lines.end(), SortGroups);
}

std::vector<WCombatDiagram::Item>::const_iterator
WCombatDiagram::findObject(bool planet, int id) const
{
    std::vector<Item>::const_iterator i = lines.begin();
    while (i != lines.end() && (i->planet != planet || i->id != id))
        ++i;
    return i;
}

void
WCombatDiagram::renameGroup(int from, int to)
{
    for (std::vector<Item>::iterator i = lines.begin(); i != lines.end(); ++i)
        if (i->group_id == from)
            i->group_id = to;
}

int
WCombatDiagram::getNumLines() const
{
    return lines.size();
}

const WCombatDiagram::Item&
WCombatDiagram::getLine(int i) const
{
    return lines[i];
}

int
WCombatDiagram::findObjectLine(bool planet, int id) const
{
    return findObject(planet, id) - lines.begin();
}

/************************** WCombatDiagramWidget *************************/

WCombatDiagramWidget::WCombatDiagramWidget(const GClassicVcrDatabase& db, const WCombatDiagram& diag)
    : db(db),
      diag(diag)
{ }

void
WCombatDiagramWidget::drawContent(GfxCanvas& can)
{
    /* Compute layout */
    const int left       = font_ems[FONT_SMALL]*TEXT_EMS;
    const int right      = RIGHT_PX;
    const int chartXRoom = getExtent().w - left - right;
    const int chartX     = getExtent().x + left;
    const int chartY     = getExtent().y;
    const int chartYRoom = getExtent().h;
    const int nlines     = diag.getNumLines();
    const int textHeight = font_heights[FONT_SMALL];

    /* Set up */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());

    /* Labels on left side */
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.setColor(UIColor::tc_Static);
    int nextTextY = chartY;
    for (int i = 0; i < nlines; ++i) {
        int y = chartY + chartYRoom * (2*i + 1) / (2*nlines) - textHeight/2;
        if (y >= nextTextY) {
            outTextF(ctx, getExtent().x, y, left, getObjectName(i));
            nextTextY = y + textHeight;
        }
    }

    /* Draw diagram */
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    std::vector<int> lastRelX(nlines);
    std::vector<int> lastOwners(nlines);
    for (int i = 0, n = db.getCount(); i < n; ++i) {
        /* Compute positions */
        const GVcrEntry& entry = db.getEntry(i);
        const TVcr& vcr = entry.getVcr();
        const int relX = chartXRoom * (i+1) / n;
        const int leftLine = diag.findObjectLine(entry.isPlanet(false), vcr.objects[0].id);
        const int rightLine = diag.findObjectLine(entry.isPlanet(true), vcr.objects[1].id);
        const int leftY = chartY + chartYRoom * (2*leftLine + 1) / (2*nlines);
        const int rightY = chartY + chartYRoom * (2*rightLine + 1) / (2*nlines);

        /* Safety check (can't happen) */
        if (leftLine >= nlines || rightLine >= nlines)
            continue;

        /* VCR status */
        int stat = entry.getStatus();
        if (stat < 0)
            stat = 0;

        /* Vertical lines. Here, we check for unique status values. */
        if (stat == VcrPlayer::VCRS_LEFT_CAPTURED) {
            ctx.setColor(WRaceList::colors[vcr.objects[1].owner - 1]);
        } else if (stat == VcrPlayer::VCRS_RIGHT_CAPTURED) {
            ctx.setColor(WRaceList::colors[vcr.objects[0].owner - 1]);
        } else if (stat == VcrPlayer::VCRS_STALEMATE) {
            ctx.setColor(COLOR_DARK);
            ctx.setLinePattern(DOTTED_LINE);
        } else {
            ctx.setColor(COLOR_BLACK);
        }
        drawVLine(ctx, chartX + relX, leftY, rightY);
        ctx.setLinePattern(SOLID_LINE);

        /* Horizontal lines */
        ctx.setLineThickness(3);
        ctx.setColor(WRaceList::colors[vcr.objects[0].owner - 1]);
        drawHLine(ctx, chartX + lastRelX[leftLine],  leftY,  chartX + relX);
        ctx.setColor(WRaceList::colors[vcr.objects[1].owner - 1]);
        drawHLine(ctx, chartX + lastRelX[rightLine], rightY, chartX + relX);
        ctx.setLineThickness(1);

        /* Explosion markers and status update */
        lastOwners[leftLine] = vcr.objects[0].owner;
        lastOwners[rightLine] = vcr.objects[1].owner;
        if (stat & VcrPlayer::VCRS_LEFT_DESTROYED) {
            lastOwners[leftLine] = 0;
            ctx.setColor(COLOR_RED);
            drawMarker(ctx, getUserMarker(0, true), GfxPoint(chartX + relX, leftY));
            ctx.setColor(COLOR_YELLOW);
            drawMarker(ctx, getUserMarker(2, true), GfxPoint(chartX + relX, leftY));
        }
        if (stat & VcrPlayer::VCRS_RIGHT_DESTROYED) {
            lastOwners[rightLine] = 0;
            ctx.setColor(COLOR_RED);
            drawMarker(ctx, getUserMarker(0, true), GfxPoint(chartX + relX, rightY));
            ctx.setColor(COLOR_YELLOW);
            drawMarker(ctx, getUserMarker(2, true), GfxPoint(chartX + relX, rightY));
        }
        if (stat & VcrPlayer::VCRS_LEFT_CAPTURED) {
            lastOwners[leftLine] = vcr.objects[1].owner;
        }
        if (stat & VcrPlayer::VCRS_RIGHT_CAPTURED) {
            lastOwners[rightLine] = vcr.objects[0].owner;
        }

        /* Remember new X */
        lastRelX[leftLine] = relX;
        lastRelX[rightLine] = relX;
    }

    /* Final lines on right side */
    ctx.setLineThickness(3);
    for (int i = 0; i < nlines; ++i) {
        if (lastOwners[i] != 0) {
            const int y = chartY + chartYRoom * (2*i + 1) / (2*nlines);
            ctx.setColor(WRaceList::colors[lastOwners[i] - 1]);
            drawHLine(ctx, chartX + lastRelX[i], y, getExtent().getRightX()-1);
        }
    }
    ctx.setLineThickness(1);
}

void
WCombatDiagramWidget::getLayoutInfo(LayoutInfo& info)
{
    int addX = font_ems[FONT_SMALL]*TEXT_EMS + RIGHT_PX;
    info.min_size  = GfxPoint(db.getCount()*5 + addX,  diag.getNumLines()*2);
    info.pref_size = GfxPoint(db.getCount()*20 + addX, diag.getNumLines()*10);
    info.flags     = li_Flexible;
}

string_t
WCombatDiagramWidget::getObjectName(int line) const
{
    const WCombatDiagram::Item& it = diag.getLine(line);
    const GVcrEntry& entry = db.getEntry(it.appears_in);

    bool side = (entry.isPlanet(true) == it.planet && entry.getVcr().objects[1].id == it.id);
    if (it.planet) {
        return format(_("%s (planet #%d)"),
                      convertGameToUtf8(getBasicString(entry.getVcr().objects[side].name)),
                      it.id);
    } else {
        return format(_("%s (ship #%d)"),
                      convertGameToUtf8(getBasicString(entry.getVcr().objects[side].name)),
                      it.id);
    }
}
