/**
  *  \file game/msg-filter.cc
  *  \brief Implementation of GMessageFilter
  *  \author Stefan Reuther
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/msg.h"

/*! \class GMessageFilter
    \brief Message filter (Killfile)

    A GMessageFilter object contains a list of filtered message
    subjects.

    \todo save and load */

/** Constructor. Makes a blank filter. */
GMessageFilter::GMessageFilter()
{ }

/** Destructor. */
GMessageFilter::~GMessageFilter()
{ }

/** Get entry for one item.
    \param header Subject line to search for
    \return i, such that *i == header. Returns filter.end() if no such object. */
GMessageFilter::entryptr_t
GMessageFilter::getEntry(const string_t& header)
{
    entryptr_t i = filter.begin();
    while (i != filter.end() && *i != header)
        ++i;
    return i;
}

/** Check whether a subject is filtered.
    \param header Subject line to search for
    \return true iff filtered */
bool
GMessageFilter::isFiltered(const string_t& header) const
{
    return const_cast<GMessageFilter*>(this)->getEntry(header)
        != const_cast<GMessageFilter*>(this)->filter.end();
}

/** Toggle whether a subject is filtered. Adds line to filter if it is not
    yet there, removes it otherwise.
    \param header Subject line to search for
    \return true if item is now filtered, false if not. */
bool
GMessageFilter::toggleFilter(const string_t& header)
{
    entryptr_t p = getEntry(header);
    if (p == filter.end()) {
        filter.push_back(header);
        return true;
    } else {
        filter.erase(p);
        return false;
    }
}

/** Add item to filter. If a duplicate is attempted to be added,
    nothing happens. */
void
GMessageFilter::addToFilter(const string_t& header)
{
    entryptr_t p = getEntry(header);
    if (p == filter.end())
        filter.push_back(header);
}

/** Remove item from filter. If that item is not filtered, nothing happens. */
void
GMessageFilter::removeFromFilter(const string_t& header)
{
    entryptr_t p = getEntry(header);
    if (p != filter.end())
        filter.erase(p);
}

/** Clear message filter. */
void
GMessageFilter::clear()
{
    filter.clear();
}

/** Check for emptiness. */
bool
GMessageFilter::empty() const
{
    return filter.empty();
}

/** Get first element. Note that the message filter does not maintain any particular order. */
GMessageFilter::iterator
GMessageFilter::begin() const
{
    return filter.begin();
}

/** Get one-past-last element. Note that the message filter does not maintain any particular order. */
GMessageFilter::iterator
GMessageFilter::end() const
{
    return filter.end();
}
