/**
  *  \file game/planet.cc
  *  \brief Planet / Starbase Management
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/planet.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/univ.h"
#include "game/planet-int.h"
#include "game/ship.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/host.h"
#include "game/team.h"
#include "util/recode.h"

static const TPlanet blank_planet = {
    UNKNOWNP,                               // owner
    UNKNOWNP,                               // planet_id
    { UNKNOWNC, UNKNOWNC, UNKNOWNC },       // fc
    UNKNOWNP, UNKNOWNP, UNKNOWNP,           // mi, fa, de
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ore mined
    UNKNOWNP,                               // colonists
    UNKNOWNP, UNKNOWNP,                     // sup, mc
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ground ore
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ore density
    UNKNOWNP, UNKNOWNP,                     // ctax, ntax
    UNKNOWNS, UNKNOWNS,                     // chap, nhap
    UNKNOWNP, UNKNOWNP, UNKNOWNP,           // ngov, natives, nrace
    UNKNOWNP,                               // temp
    UNKNOWNP                                // base
};

static const TAutobuildSettings standard_autobuild = {
    { 1000, 1000, 1000, 20 },
    { 5, 10, 3, 2 }
};

GPlanet::MapInfo::MapInfo(int id)
    : id(id), position(), known_to_not_exist(false), name("?"), flags(0),
      autobuild(standard_autobuild)
{ }

GPlanet::MapInfo::~MapInfo()
{ }

GPlanet::PlanetInfo::PlanetInfo(int id)
{
    data = blank_planet;
    data.planet_id = id;
    known_to_have_natives = false;
    time[ts_Minerals] = time[ts_Colonists] = time[ts_Natives] = time[ts_Cash] = 0;
    industry_level = mp16_t();
}

/******************************** GPlanet ********************************/

/** Constructor. Create blank planet of given Id.
    \param id Planet Id */
GPlanet::GPlanet(int id)
    : planet_info(0), base_info(0), map_info(id), planet_src_flags(), base_src_flags()
{ }

GPlanet::GPlanet(const GPlanet& p)
    : GMapObject(),
      unit_scores(p.unit_scores),
      planet_info(0), base_info(0),
      map_info(p.map_info),
      planet_src_flags(p.planet_src_flags),
      base_src_flags(p.base_src_flags)
{
    if (p.planet_info)
        planet_info = new PlanetInfo(*p.planet_info);
    if (p.base_info)
        base_info = new BaseInfo(*p.base_info);
    setIsMarked(p.isMarked());
    setIsPlayed(p.isPlayed());
}

GPlanet::~GPlanet()
{
    delete planet_info;
    delete base_info;
}

/** Create planet info record. */
GPlanet::PlanetInfo&
GPlanet::createPlanetInfo()
{
    if (!planet_info)
        planet_info = new PlanetInfo(map_info.id);
    return *planet_info;
}

/** Get data to be used for display.
    \returns planet data record */
const TPlanet&
GPlanet::getDisplayedPlanet() const
{
    return planet_info ? planet_info->data : blank_planet;
}

/** \name Loading */
//@{

/** Add .dat file entry. This assumes that when we see a planet
    through several .dat files, we get the same one each time.
    \param dat   parsed .dat file entry
    \param srcf  source flag value to use for this entry */
void
GPlanet::addPlanetData(const TPlanet& dat, GPlayerSet srcf)
{
    // FIXME: older PHost versions clear the following fields of a
    // planet when sending a pdata planet target for exploration of an
    // unowned planet (owned planets never generate a target):
    // colonists, supplies, credits, mines, factories, defense,
    // coltax, nattax. We may want to merge that information somehow,
    // or detect and/or merge it.
    createPlanetInfo().data = dat;
    planet_src_flags |= srcf;
}

/** Add starbase .dat file entry. See addPlanetData() for more
    information. */
void
GPlanet::addBaseData(const TStarbase& bdat, GPlayerSet srcf)
{
    if (!base_info) {
        base_info = new BaseInfo;
        base_info->data = bdat;
    }
    base_src_flags |= srcf;
}

/** Add history file entry. */
void
GPlanet::addHistoryData(const TDbPlanet& data)
{
    if (planet_src_flags.empty()) {
        // Accept only when not scanned otherwise
        PlanetInfo& pi = createPlanetInfo();

        // This assumes only one history data source. Otherwise,
        // we have to merge.
        pi.data               = data.planet;
        pi.time[ts_Minerals]  = data.time[ts_Minerals];
        pi.time[ts_Colonists] = data.time[ts_Colonists];
        pi.time[ts_Natives]   = data.time[ts_Natives];
        pi.time[ts_Cash]      = data.time[ts_Cash];
        if (data.planet.factories >= 30000) {
            pi.data.factories = mp16_t().getRawValue();
            pi.industry_level = mp16_t(data.planet.factories - 30000);
        } else {
            pi.industry_level = mp16_t();
        }
        pi.known_to_have_natives = (data.known_to_have_natives != 0);
    }
}

/** Set planet position. */
void
GPlanet::setXY(int ax, int ay)
{
    map_info.position.x = ax;
    map_info.position.y = ay;
}

/** Set planet name. */
void
GPlanet::setName(const string_t& aname)
{
    map_info.name = aname;
}

/** Set whether non-existance of this planet is known. There is no way
    to explicitly specify that a planet does not exist. To build maps
    with fewer than 500 planets, people move planets to odd positions.
    Recent PHosts send a util.dat message whenever they consider a
    planet to be non-existant, to make sure that the clients' idea of
    which planets do exist agrees with PHost's. */
void
GPlanet::setKnownToNotExist(bool value)
{
    map_info.known_to_not_exist = value;
}

/** Add autobuild record from database file.
    \param abs [in] Autobuild record */
void
GPlanet::setAutobuild(const TAutobuildSettings& abs)
{
    map_info.autobuild = abs;
}

/** Get planet data record for storage in data files.
    \param dat [out] planet data */
void
GPlanet::getPlanetData(TPlanet& dat) const
{
    if (planet_info)
        dat = planet_info->data;
    else
        dat = blank_planet;
}

/** Get starbase data record for storage in data files.
    \param bdat [out] base data
    \pre hasBaseData() */
void
GPlanet::getBaseData(TStarbase& bdat) const
{
    ASSERT(base_info);
    bdat = base_info->data;
}

/** Get Autobuild record for storage in database file.
    \param abs [out] Autobuild record */
void
GPlanet::getAutobuild(TAutobuildSettings& abs) const
{
    abs = map_info.autobuild;
}

/** Get planet history record for storage in database file.
    \pre hasAnyPlanetData()
    \param data [out] History record */
void
GPlanet::getHistoryData(TDbPlanet& data) const
{
    ASSERT(planet_info);
    data.planet             = planet_info->data;
    data.time[ts_Minerals]  = planet_info->time[ts_Minerals];
    data.time[ts_Colonists] = planet_info->time[ts_Colonists];
    data.time[ts_Natives]   = planet_info->time[ts_Natives];
    data.time[ts_Cash]      = planet_info->time[ts_Cash];
    data.known_to_have_natives = planet_info->known_to_have_natives;
    if (planet_info->industry_level.isKnown() && mp16_t(planet_info->data.factories).isKnown()) {
        data.planet.factories = planet_info->industry_level + 30000;
    }
    if (base_info) {
        data.planet.build_base = 1;
    }
}

//@}

//! \name Checking
//@{

/** Do internal checks for this planet. Internal checks do not require
    a partner to interact with. This will fix the problems, and
    display appropriate messages. */
void
GPlanet::internalCheck()
{
    map_info.flags = 0;

    /* set existance flag according to x/y */
    if (isClippedCoord(map_info.position) || map_info.known_to_not_exist) {
        map_info.flags &= ~FlagPlanetExists;
    } else {
        map_info.flags |= FlagPlanetExists;
    }

    if (hasPlanetData()) {
        /* we have a DAT entry. This planet MUST exist. */
        if (!(map_info.flags & FlagPlanetExists)) {
            const char* fmt = (map_info.known_to_not_exist
                               ? _("Planet #%d has data although it is reported as non-existant -- host confused?")
                               : _("Planet #%d exists for Host, but is outside valid range"));
            console.write(LOG_WARN, format(fmt, map_info.id));
            map_info.flags |= FlagPlanetExists;
        }
    }

    if (hasBaseData()) {
        /* we have a BDATA entry. There must also be a PDATA entry */
        // FIXME this will make PCC2 write invalid files, i.e. where
        // PDATA.DAT and PDATA.DIS disagree.
        if (!hasPlanetData()) {
            console.write(LOG_WARN,
                          format(_("Starbase #%d does not have a planet -- deleting it"), map_info.id));
            delete base_info;
            base_info = 0;
            base_src_flags = GPlayerSet();
        }
    }
}

/** Combined checks, phase 2. This will do all post-processing which
    needs a partner to interact with. */
void
GPlanet::combinedCheck2(GUniverse& univ)
{
    if (hasPlanetData()) {
        /* we've seen this planet this turn */
        planet_info->time[ts_Minerals]
            = planet_info->time[ts_Colonists]
            = planet_info->time[ts_Natives]
            = planet_info->time[ts_Cash]
            = univ.getTurnNumber();
    } else if (hasAnyPlanetData()) {
        /* we have taken this planet from the history */
        if (isOwnerKnown() && univ.isPlaying(getOwner())) {
            /* planet is played by us, but we do no longer own it */
            planet_info->data.owner = 0;
        }
    }
}
//@}

//! @name Accessors
//@{

/** True iff this planet has a starbase. */
bool
GPlanet::hasBase() const
{
    return isPlayed() ? base_info != 0 : planet_info && planet_info->data.build_base > 0;
}

/** Get history timestamp. */
int16
GPlanet::getHistoryTimestamp(TDbPlanetTimestamp which_one) const
{
    if (hasAnyPlanetData())
        return planet_info->time[which_one];
    else
        return 0;
}

/** Clear orbit flags. This will remove all planet ship marks. */
void
GPlanet::clearOrbitFlags()
{
    map_info.flags &= ~(FlagOwnShipsInOrbit | FlagEnemyShipsInOrbit | FlagGuessedEnemyInOrbit
                        | FlagAlliedShipsInOrbit | FlagGuessedAllyInOrbit);
}

/** Notify planet of ship in orbit. This will compute the planet markings. */
void
GPlanet::setShipInOrbit(const GShip& s)
{
    switch (getPlayerRelation(s.getOwner())) {
     case is_Me:
        map_info.flags |= FlagOwnShipsInOrbit;
        break;
     case is_Enemy:
        if (s.isReliablyVisible())
            map_info.flags |= FlagEnemyShipsInOrbit;
        else
            map_info.flags |= FlagGuessedEnemyInOrbit;
        break;
     case is_Ally:
        if (s.isReliablyVisible())
            map_info.flags |= FlagAlliedShipsInOrbit;
        else
            map_info.flags |= FlagGuessedAllyInOrbit;
        break;
    }
}

/** Check whether specified coordinates are clipped (i.e.\ nonexistant
    planet). There is no way for clients to tell that a planet is
    non-existant; the ad-hoc solution is moving planets far away.
    Consistent with PCC 1.x handling of this. */
bool
GPlanet::isClippedCoord(GPoint pt)
{
    // FIXME: this routine does not belong into GPlanet, but into
    // GChartConfiguration or similar
    if (pt.x <= 0 || pt.y <= 0 || pt.x > 9000 || pt.y > 9000 || !pt.isOnMap())
        return true;
    else
        return false;
}

string_t
GPlanet::getName() const
{
    return map_info.name;
}

string_t
GPlanet::getFullName() const
{
    return format(_("Planet #%d: %s"), map_info.id, map_info.name);
}

int
GPlanet::getId() const
{
    return map_info.id;
}

int
GPlanet::getOwner() const
{
    if (planet_info && mp16_t(planet_info->data.owner).isKnown())
        return planet_info->data.owner;
    else
        return 0;
}

bool
GPlanet::isExistant() const
{
    return map_info.flags & FlagPlanetExists;
}

GPoint
GPlanet::getPos() const
{
    return map_info.position;
}

/* Accessors: structures */

/** Get number of structures.
    \param ps Which structure to query
    \return number of structures, or unknown */
mp16_t
GPlanet::getStructures(TPlanetaryStructure ps) const
{
    if (!planet_info)
        return mp16_t();

    switch (ps) {
     case str_Mines:
        return mp16_t(planet_info->data.mines);
     case str_Factories:
        return mp16_t(planet_info->data.factories);
     case str_Defense:
        return mp16_t(planet_info->data.defense);
     case str_BaseDefense:
        // FIXME: if we do not know whether there is a base, the right
        // answer would be mp16_t().
        return base_info ? mp16_t(base_info->data.defense) : mp16_t(0);
    }
    return mp16_t();
}

/** Set number of structures. Works for own and scanned foreign planets.
    This cannot set base defense if the planet doesn't have a base.

    \param ps    Which structure to set
    \param count New amount */
void
GPlanet::setStructures(TPlanetaryStructure ps, mp16_t count)
{
    /* is there a change? */
    if (count == getStructures(ps))
        return;

    switch (ps) {
     case str_Mines:
        createPlanetInfo().data.mines = count.getRawValue();
        break;
     case str_Factories:
        createPlanetInfo().data.factories = count.getRawValue();
        break;
     case str_Defense:
        createPlanetInfo().data.defense = count.getRawValue();
        break;
     case str_BaseDefense:
        ASSERT(base_info);
        base_info->data.defense = count.getRawValue();
        break;
    }
    markDirty();
}

/** Get maximum allowed structures of a type.
    \param ps Structure to query */
mp32_t
GPlanet::getMaxStructures(TPlanetaryStructure ps) const
{
    if (mp32_t(getDisplayedPlanet().colonists).isKnown() && isOwnerKnown())
        return mp32_t(getMaxStructures(ps, getColonists()));
    else
        return mp32_t();
}

/** Get industry level of this planet. Reports the industry level from
    known structure counts if available, otherwise from sensor scans. */
mp16_t
GPlanet::getIndustryLevel() const
{
    mp16_t mi = getStructures(str_Mines);
    mp16_t fa = getStructures(str_Factories);
    if (mi.isKnown() && fa.isKnown())
        return mp16_t(getIndustryLevel(mi + fa));
    else if (planet_info)
        return planet_info->industry_level;
    else
        return mp16_t();
}

/** Get industry level for a given structure count.
    \param mifa Mines+Factories
    \return level (ind_XXX) */
int
GPlanet::getIndustryLevel(int mifa) const
{
    if (host.isPHost()) {
        int level = mifa / 30;
        if (level > ind_Heavy)
            return ind_Heavy;
        else
            return level;
    } else {
        if (mifa >= 100)
            return ind_Heavy;
        else
            return ind_Light;
    }
}

/** Set industry level for this planet. This routine only makes sense
    for planets we do not play. */
void
GPlanet::setIndustryLevel(mp16_t level)
{
    // If we're playing this planet, setting the industry level is useless
    if (hasPlanetData())
        return;

    // We're not playing the planet, so set the level and check for
    // conflicts with our other data
    createPlanetInfo().industry_level = level;
    if (level.isKnown()) {
        mp16_t mi = getStructures(str_Mines);
        mp16_t fa = getStructures(str_Factories);
        if (mi.isKnown() && fa.isKnown()) {
            // We know mines and factories. Does this conflict with our level?
            if (getIndustryLevel(mi + fa) != level) {
                // yes, so forget mi+fa
                setStructures(str_Mines, mp16_t());
                setStructures(str_Factories, mp16_t());
            }
        } else if (mi.isKnown() || fa.isKnown()) {
            // We know mines or factories, but not both. Does this conflict with our level?
            int mifa = 0;
            if (mi.isKnown())
                mifa += mi;
            if (fa.isKnown())
                mifa += fa;
            if (getIndustryLevel(mifa) > level) {
                // our stored mine/factory count would yield a larger
                // level than reported. This means our stored mi/fa is
                // outdated.
                setStructures(str_Mines, mp16_t());
                setStructures(str_Factories, mp16_t());
            }
        }
    }
}

/* Accessors: colonists */

mn16_t
GPlanet::getColonistHappiness() const
{
    return mn16_t(getDisplayedPlanet().colonist_happy);
}

void
GPlanet::setColonistHappiness(mn16_t happiness)
{
    createPlanetInfo().data.colonist_happy = happiness.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getColonistTax() const
{
    return mp16_t(getDisplayedPlanet().colonist_tax);
}

void
GPlanet::setColonistTax(mp16_t tax)
{
    createPlanetInfo().data.colonist_tax = tax.getRawValue();
    markDirty();
}



mn32_t
GPlanet::getColonists() const
{
    // FIXME: this uses mn32_t because the reservation might exceed the amount(?)
    // if we do not know the amount, assume the reservation is empty
    uint32 raw_value = getDisplayedPlanet().colonists;
    if (!mp32_t(raw_value).isKnown())
        return mn32_t();
    else
        return mn32_t(raw_value - cargo.getReservedCargo(el_Colonists));
}

void
GPlanet::setColonists(mn32_t clans)
{
    // FIXME: must handle reservations!
    if (clans != getColonists())
        return;
    createPlanetInfo().data.colonists = clans.getRawValue();
    markDirty();
}

bool
GPlanet::isOwnerKnown() const
{
    return mp16_t(getDisplayedPlanet().owner).isKnown();
}

/* Accessors: Natives */

mp16_t
GPlanet::getNativeGovernment() const
{
    return mp16_t(getDisplayedPlanet().native_gov);
}

void
GPlanet::setNativeGovernment(mp16_t gov)
{
    createPlanetInfo().data.native_gov = gov.getRawValue();
    markDirty();
}

mn16_t
GPlanet::getNativeHappiness() const
{
    return mn16_t(getDisplayedPlanet().native_happy);
}

void
GPlanet::setNativeHappiness(mn16_t happiness)
{
    createPlanetInfo().data.native_happy = happiness.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getNativeRace() const
{
    return mp16_t(getDisplayedPlanet().native_race);
}

void
GPlanet::setNativeRace(mp16_t race)
{
    createPlanetInfo().data.native_race = race.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getNativeTax() const
{
    return mp16_t(getDisplayedPlanet().native_tax);
}

void
GPlanet::setNativeTax(mp16_t tax)
{
    createPlanetInfo().data.native_tax = tax.getRawValue();
    markDirty();
}

mp32_t
GPlanet::getNatives() const
{
    return mp32_t(getDisplayedPlanet().natives);
}

void
GPlanet::setNatives(mp32_t natives)
{
    createPlanetInfo().data.natives = natives.getRawValue();
    markDirty();
}

bool
GPlanet::isKnownToHaveNatives() const
{
    /* We claim the planet has natives if
       ...we know that it has some from RGA        -or-
       ...we know the race (bioscan)               -or-
       ...we know the population (pillage) */
    return planet_info
        && (planet_info->known_to_have_natives
            || (mp16_t(planet_info->data.native_race).isKnown()
                && planet_info->data.native_race != nr_None)
            || (mp32_t(planet_info->data.natives).isKnown()
                && planet_info->data.natives != 0));
}

void
GPlanet::setKnownToHaveNatives(bool known)
{
    createPlanetInfo().known_to_have_natives = known;
    markDirty();
}

/* Accessors: Fcode */

string_t
GPlanet::getFCode() const
{
    if (planet_info)
        return convertGameToUtf8(getBasicString(planet_info->data.fcode));
    else
        return string_t();
}

void
GPlanet::setFCode(const string_t& str)
{
    storeBasicString(createPlanetInfo().data.fcode, convertUtf8ToGame(str));
    markDirty();
}

bool
GPlanet::isFCodeKnown() const
{
    return planet_info
        && (planet_info->data.fcode[0] != UNKNOWNC
            || planet_info->data.fcode[1] != UNKNOWNC
            || planet_info->data.fcode[2] != UNKNOWNC);
}

/* Accessors: base building */

bool
GPlanet::isBuildingBase() const
{
    return hasPlanetData() && planet_info->data.build_base;
}

void
GPlanet::setBuildBaseFlag(bool b)
{
    ASSERT(hasPlanetData());
    if (b != isBuildingBase()) {
        planet_info->data.build_base = b;
        markDirty();
    }
}

/* Accessors: Environment */

mp16_t
GPlanet::getOreDensity(int index) const
{
    return mp16_t(getDisplayedPlanet().ore_density[index]);
}

void
GPlanet::setOreDensity(int index, mp16_t amount)
{
    createPlanetInfo().data.ore_density[index] = amount;
    markDirty();
}

mp32_t
GPlanet::getOreGround(int index) const
{
    return mp32_t(getDisplayedPlanet().ore_ground[index]);
}

void
GPlanet::setOreGround(int index, mp32_t amount)
{
    createPlanetInfo().data.ore_ground[index] = amount.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getTemperature() const
{
    mp16_t temp_code = mp16_t(getDisplayedPlanet().temp_code);
    if (temp_code.isKnown())
        return mp16_t(100 - temp_code);
    else
        return temp_code;
}

void
GPlanet::setTemperature(mp16_t temp)
{
    if (temp.isKnown())
        createPlanetInfo().data.temp_code = 100 - temp;
    else if (planet_info)
        planet_info->data.temp_code = mp16_t();
    markDirty();
}

/* Accessors: cargo */

int32
GPlanet::getMoney() const
{
    // FIXME: handle partial knowledge?
    return planet_info->data.money - cargo.getReservedCargo(el_Money);
}

int32
GPlanet::getSupplies() const
{
    // FIXME: handle partial knowledge?
    return planet_info->data.supplies - cargo.getReservedCargo(el_Supplies);
}

/** Get cargo of a particular type, as to be used by most code.
    This will remove reservations from the amount reported. */
int32
GPlanet::getCargo(GCargoType type) const
{
    // FIXME: handle partial knowledge?
    return getCargoRaw(type) - cargo.getReservedCargo(type);
}

/** Get cargo, as it is in underlying data structure. */
int32
GPlanet::getCargoRaw(GCargoType type) const
{
    // FIXME: handle partial knowledge?
    const TPlanet& dpy = getDisplayedPlanet();
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return dpy.ore_mined[type];
     case el_Supplies:
        return dpy.supplies;
     case el_Money:
        return dpy.money;
     case el_Fighters:
        if (base_info)
            return base_info->data.ammo_store[10];
        return 0;
     case el_Colonists:
        return dpy.colonists;
     default:
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        if (base_info)
            return base_info->data.ammo_store[type - el_Torps];
        return 0;
    }
}

void
GPlanet::changeCargoRaw(GCargoType type, int32 amount)
{
    if (!amount)
        return;

    // FIXME
    ASSERT(hasPlanetData());
    markDirty();

    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        planet_info->data.ore_mined[type] += amount;
        return;
     case el_Supplies:
        planet_info->data.supplies += amount;
        return;
     case el_Money:
        planet_info->data.money += amount;
        return;
     case el_Fighters:
        ASSERT(base_info);
        base_info->data.ammo_store[10] += amount;
        return;
     case el_Colonists:
        planet_info->data.colonists += amount;
        return;
     default:
        ASSERT(base_info);
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        base_info->data.ammo_store[type - el_Torps] += amount;
        return;
    }
}

// FIXME: this function should go
bool
GPlanet::isCargoAmountKnown(GCargoType type) const
{
    if (!planet_info)
        return false;
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return planet_info->data.ore_mined[type] != UNKNOWNP;
     case el_Supplies:
        return planet_info->data.supplies != UNKNOWNP;
     case el_Money:
        return planet_info->data.money != UNKNOWNP;
     case el_Fighters:
        return base_info != 0;
     case el_Colonists:
        return planet_info->data.colonists != UNKNOWNP;
     default:
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        return base_info != 0;
    }
}

/* Accessors: base specs */

int
GPlanet::getBaseDamage() const
{
    return base_info->data.damage;
}

int
GPlanet::getBaseMission() const
{
    return base_info->data.mission;
}

void
GPlanet::setBaseMission(int mission)
{
    ASSERT(base_info);
    if (mission != getBaseMission()) {
        base_info->data.mission = mission;
        markDirty();
    }
}

int
GPlanet::getBaseShipyardOrder() const
{
    return base_info->data.shipyard_action;
}

int
GPlanet::getBaseShipyardId() const
{
    return base_info->data.shipyard_id;
}


void
GPlanet::setBaseShipyardOrder(int action, int id)
{
    ASSERT(base_info);
    if (action != getBaseShipyardOrder() || id != getBaseShipyardId()) {
        base_info->data.shipyard_action = action;
        base_info->data.shipyard_id     = id;
        markDirty();
    }
}


int
GPlanet::getBaseTech(TTechLevel level) const
{
    return base_info->data.tech_levels[level];
}

void
GPlanet::setBaseTech(TTechLevel level, int value)
{
    ASSERT(value > 0 && value <= 10);
    ASSERT(base_info);
    if (value != base_info->data.tech_levels[level]) {
        base_info->data.tech_levels[level] = value;
        markDirty();
    }
}

int
GPlanet::getBaseBuildHull() const
{
    ASSERT(base_info);
    if (base_info->data.build.hull)
        return getTruehull(base_info->data.owner, base_info->data.build.hull);
    else
        return 0;
}

/** Get base build order. Note that the build order is normalized to contain
    the real hull Id, not the slot number. */
TBuildOrder
GPlanet::getBaseBuildOrder() const
{
    TBuildOrder bo = base_info->data.build;
    bo.hull = getBaseBuildHull();
    return bo;
}

/** Set base build order. Note that the build order is normalized to contain
    the real hull Id, not the slot number. */
void
GPlanet::setBaseBuildOrder(const TBuildOrder& order)
{
    if (!order.hull) {
        base_info->data.build.hull = base_info->data.build.engine =
            base_info->data.build.beam_type = base_info->data.build.beam_count =
            base_info->data.build.torp_type = base_info->data.build.torp_count =
            base_info->data.build.unused = 0;
    } else {
        int slot = getTruehullSlot(base_info->data.owner, order.hull);
        ASSERT(slot);
        base_info->data.build = order;
        base_info->data.build.hull = slot;
    }
    markDirty();
}

/** Get hull slot used for ship being built. Whereas the above functions deal
    with real hull numbers, this one returns a slot; this simplifies things
    every once in a while. */
int
GPlanet::getBaseBuildOrderHullSlot() const
{
    return base_info->data.build.hull;
}

/** Get ship which is trying to clone at this planet. This runs
    through the ship list, so use with care.
    \param univ universe which contains the ships */
int
GPlanet::getShipCloningHere(const GUniverse& univ) const
{
    // FIXME: check whether 'cln' works
    if (!config.AllowShipCloning())
        return 0;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& s = univ.getShip(i);
        if (s.isExistant() && s.isCloningAt(*this))
            return i;
    }
    return 0;
}

int
GPlanet::getBaseEngineStore(int slot) const
{
    return base_info->data.engine_store[slot-1];
}

int
GPlanet::getBaseBeamStore(int slot) const
{
    return base_info->data.beam_store[slot-1];
}

int
GPlanet::getBaseHullStoreSlot(int slot) const
{
    return base_info->data.hull_store[slot-1];
}

int
GPlanet::getBaseLauncherStore(int slot) const
{
    return base_info->data.launcher_store[slot-1];
}

/** Get amount of ammunition. This is a convenience function to simplify implementation
    of the script interface.
    \param slot [1,11], where [1,10] are the torps, 11 is fighters */
int
GPlanet::getBaseAmmoStore(int slot) const
{
    ASSERT(slot > 0 && slot <= NUM_TORPS+1);
    if (slot > 0 && slot <= NUM_TORPS)
        return getCargo(getCargoTypeFromTorpType(slot));
    else
        return getCargo(el_Fighters);
}

/** Set number of engines in starbase storage. */
void
GPlanet::setBaseEngineStore(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_ENGINES);
    base_info->data.engine_store[slot-1] = amount;
    markDirty();
}

/** Set number of beams in starbase storage. */
void
GPlanet::setBaseBeamStore(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_BEAMS);
    base_info->data.beam_store[slot-1] = amount;
    markDirty();
}

/** Set number of hulls in starbase storage. Note that this takes a
    truehull slot number, not a real hull Id. */
void
GPlanet::setBaseHullStoreSlot(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_HULLS_PER_PLAYER);
    base_info->data.hull_store[slot-1] = amount;
    markDirty();
}

/** Set number of torpedo launchers in starbase storage. */
void
GPlanet::setBaseLauncherStore(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_TORPS);
    base_info->data.launcher_store[slot-1] = amount;
    markDirty();
}

/** Get autobuild goal for a structure. Known for all planets. */
int32
GPlanet::getAutobuildGoal(TPlanetaryStructure ps) const
{
    return map_info.autobuild.goal[ps];
}

/** Set autobuild goal for a structure. */
void
GPlanet::setAutobuildGoal(TPlanetaryStructure ps, int32 value)
{
    if (value != map_info.autobuild.goal[ps]) {
        map_info.autobuild.goal[ps] = value;
        markDirty();
    }
}

/** Get autobuild speed for a structure. Known for all planets. */
int
GPlanet::getAutobuildSpeed(TPlanetaryStructure ps) const
{
    return map_info.autobuild.speed[ps];
}

/** Set autobuild speed for a structure. */
void
GPlanet::setAutobuildSpeed(TPlanetaryStructure ps, int value)
{
    if (value != map_info.autobuild.speed[ps]) {
        map_info.autobuild.speed[ps] = value;
        markDirty();
    }
}

/** Get name of a planetary structure. */
const char*
GPlanet::getStructureName(TPlanetaryStructure ps)
{
    static const char*const names[] = {
        N_("Mineral Mines"),
        N_("Factories"),
        N_("Defense Posts"),
        N_("Starbase Defense")
    };
    return _(names[ps]);
}

/****************************** GPlanetType ******************************/

/** Constructor.
    \param trn Turn object through which we iterate. */
GPlanetType::GPlanetType()
    : GIndexedObjectType(NUM_PLANETS)
{ }

bool
GPlanetType::isExistant(int index)
{
    GUniverse* u = getUniverse();
    if (u != 0 && index > 0 && index <= NUM_PLANETS) {
        const GPlanet& p = u->getPlanet(index);
        if (p.isExistant() && p.isPlayed() && p.hasPlanetData())
            return true;
    }
    return false;
}

GPlanet&
GPlanetType::getObject(int index)
{
    return getUniverse()->getPlanet(index);
}

/***************************** GAnyPlanetType ****************************/

/** Constructor.
    \param trn Turn object through which we iterate. */
GAnyPlanetType::GAnyPlanetType()
    : GIndexedObjectType(NUM_PLANETS)
{ }

bool
GAnyPlanetType::isExistant(int index)
{
    GUniverse* u = getUniverse();
    return u != 0
        && u->isValidPlanetId(index)
        && u->getPlanet(index).isExistant();
}

GPlanet&
GAnyPlanetType::getObject(int index)
{
    return getUniverse()->getPlanet(index);
}

/******************************* GBaseType *******************************/

GBaseType::GBaseType()
    : GIndexedObjectType(NUM_PLANETS)
{ }

bool
GBaseType::isExistant(int index)
{
    GUniverse* u = getUniverse();
    if (u != 0 && index > 0 && index <= NUM_PLANETS) {
        const GPlanet& p = u->getPlanet(index);
        if (p.isExistant() && p.isPlayed() && p.hasBase())
            return true;
    }
    return false;
}

GPlanet&
GBaseType::getObject(int index)
{
    return getUniverse()->getPlanet(index);
}

/**************************** GPlanetTransfer ****************************/

GPlanetTransfer::GPlanetTransfer(GPlanet& planet)
    : planet(planet)
{
    setFlags(can_GetUnload | can_Sup2mc);
    planet.getCargoArbiter().incUseCount();
    if (!planet.isPlayed())
        throw GError(GError::eNotPlaying, _("You do not own this planet"));
}

GPlanetTransfer::~GPlanetTransfer()
{
    planet.getCargoArbiter().decUseCount();
    if (!isFinished())
        doCancel();
}

int32
GPlanetTransfer::getCargoOnObject(GCargoType type)
{
    return planet.getCargo(type);
}

void
GPlanetTransfer::changeCargoOnObject(GCargoType type, int32 amount)
{
    planet.changeCargoRaw(type, amount);
}

void
GPlanetTransfer::changeReservedCargo(GCargoType type, int32 amount)
{
    planet.getCargoArbiter().changeReservedCargo(type, amount);
}

bool
GPlanetTransfer::canHaveCargo(GCargoType cargo_id)
{
    if (cargo_id == el_Fighters
        || (cargo_id >= el_Torps && cargo_id < el_Torps + NUM_TORPS))
    {
        if (!planet.hasBase())
            return false;
    }
    return true;
}

int32
GPlanetTransfer::getMaxCargo(GCargoType cargo_id)
{
    // FIXME: Host .31 torp safety
    /* the torp/ore limits are arbitrary, but should match Tim's TRN check. */
    if (cargo_id == el_Fighters)
        return 60; /* FIXME */
    else if (cargo_id >= el_Torps && cargo_id < el_Torps + NUM_TORPS)
        return 10000;
    else
        return 1000000000;
}

string_t
GPlanetTransfer::getName() const
{
    // FIXME: PCC 1.x says "$NAME starbase" or "planet $NAME" here.
    // should we use same wording?
    return planet.getName();
}

int32
GPlanetTransfer::getDisplayOffset(GCargoType /*type*/)
{
    return 0;
}


//@}
