/**
  *  \file reshack/w32import.cc
  *  \brief Resource Editor: Win32 Font Import
  */

#include <windows.h>
#include <memory>
#include "reshack/w32import.h"
#include "ui/msgbox.h"
#include "util/nls.h"
#include "cpluslib/format.h"

#include <map>
#include <fstream>

namespace {
    struct Objects {
        HFONT hFont;
        HDC   screenDC;
        HDC   virtDC;
    };
}

static bool
createObjects(Objects& o, const LOGFONT& lf)
{
    // Clear
    o.hFont    = 0;
    o.screenDC = 0;
    o.virtDC   = 0;

    // Create font
    o.hFont = CreateFontIndirect(&lf);
    if (!o.hFont)
        return false;

    // Create DCs
    o.screenDC = CreateDC("DISPLAY", 0, 0, 0);
    if (!o.screenDC)
        return false;
    o.virtDC = CreateCompatibleDC(o.screenDC);
    if (!o.virtDC)
        return false;

    // http://support.microsoft.com/kb/306198/en-us
    HDC hScreenIC = CreateIC("DISPLAY", NULL, NULL, NULL);
    HFONT hOldFont = (HFONT) SelectObject(hScreenIC, o.hFont);
    SelectObject(hScreenIC, hOldFont);
    DeleteDC(hScreenIC);

    return true;
}

static void
destroyObjects(Objects& o)
{
    if (o.virtDC)
        DeleteDC(o.virtDC);
    if (o.screenDC)
        DeleteDC(o.screenDC);
    if (o.hFont)
        DeleteObject(o.hFont);
}

Ptr<GfxBitmapFont>
importWin32Font()
{
    // Choose a font
    LOGFONT lf;
    CHOOSEFONT cf;
    ZeroMemory(&cf, sizeof(cf));
    cf.lStructSize = sizeof(cf);
    cf.hwndOwner   = 0;
    cf.hDC         = 0;
    cf.lpLogFont   = &lf;
    cf.Flags       = CF_SCREENFONTS;
    if (!ChooseFont(&cf)) {
        return 0;
    }

    // Raise quality
    lf.lfQuality = ANTIALIASED_QUALITY;

    // Create objects
    Objects o;
    if (!createObjects(o, lf)) {
        destroyObjects(o);
        messageBox(_("Unable to create system objects."), _("Import Font"));
        return 0;
    }
    SelectObject(o.virtDC, o.hFont);
    SelectObject(o.virtDC, GetStockObject(BLACK_BRUSH));
    SelectObject(o.virtDC, GetStockObject(BLACK_PEN));

    // Create font
    Ptr<GfxBitmapFont> newFont = new GfxBitmapFont();

    // GetGlyphIndicesW exists on W2k, and is not in MinGW header files.
    typedef WINGDIAPI DWORD WINAPI (*tGetGlyphIndicesW)(IN HDC, IN LPCWSTR, IN int, OUT LPWORD, IN DWORD);
    tGetGlyphIndicesW pGetGlyphIndicesW;
    HINSTANCE hGdi = LoadLibrary("gdi32.dll");
    if (hGdi != 0)
        pGetGlyphIndicesW = (tGetGlyphIndicesW) GetProcAddress(hGdi, "GetGlyphIndicesW");
    enum { GGI_MARK_NONEXISTING_GLYPHS = 1 };

    // Do it
    int errs = 0;
    int success = 0;
    std::map<int,int> pixels;
    for (int i = 32; i < 0x3000; ++i) {
        wchar_t str = i;

        // Does this exist in the font? (and can we find out?)
        if (pGetGlyphIndicesW) {
            WORD gi;
            pGetGlyphIndicesW(o.virtDC, &str, 1, &gi, GGI_MARK_NONEXISTING_GLYPHS);
            if (gi == 0xFFFF) {
                continue;
            }
        }

        // Compute size
        RECT rect;
        rect.left = rect.top = 0;
        if (!DrawTextW(o.virtDC, &str, 1, &rect, DT_TOP | DT_LEFT | DT_CALCRECT | DT_SINGLELINE | DT_NOPREFIX)) {
            ++errs;
            continue;
        }
        if (rect.right == 0) {
            continue;
        }

        // Create bitmap to receive the character
        //HBITMAP bmp = CreateCompatibleBitmap(o.virtDC, rect.right, rect.bottom);
        HBITMAP bmp = CreateBitmap(rect.right, rect.bottom, 1, 32, 0);
        if (!bmp) {
            ++errs;
            continue;
        }
        HBITMAP prevBMP = (HBITMAP) SelectObject(o.virtDC, bmp);
        FillRect(o.virtDC, &rect, (HBRUSH) GetStockObject(WHITE_BRUSH));

        // Draw character
        if (DrawTextW(o.virtDC, &str, 1, &rect, DT_TOP | DT_LEFT | DT_SINGLELINE | DT_NOPREFIX)) {
            // Successfully drawn, so import the glyph
            std::auto_ptr<GfxBitmapGlyph> glyph(new GfxBitmapGlyph(rect.right, rect.bottom));
            for (int y = 0; y < rect.bottom; ++y) {
                for (int x = 0; x < rect.right; ++x) {
                    COLORREF color = GetPixel(o.virtDC, x, y);
                    int brightness = GetRValue(color) + GetGValue(color) + GetBValue(color);
                    ++pixels[brightness];
                    if (brightness < 0x200) {
                        // black
                        glyph->set(x, y, true);
                    } else if (brightness < 0x280) {
                        // half
                        glyph->addAAHint(x, y);
                    } else {
                        // white
                    }
                }
            }
            newFont->addNewGlyph(i, glyph.release());
            ++success;
        } else {
            ++errs;
        }

        // Clean up
        SelectObject(o.virtDC, prevBMP);
        DeleteObject(bmp);
    }
    destroyObjects(o);
    if (hGdi)
        FreeLibrary(hGdi);

    messageBox(format(_("%d characters imported successfully, %d errors."), success, errs), _("Font Import"));

    std::ofstream out("test.txt");
    for (std::map<int,int>::iterator i = pixels.begin(); i != pixels.end(); ++i)
        out << i->first << " x " << i->second << "\n";
    
    return newFont;
}
