/**
  *  \file ui/richhelp.cc
  *  \brief Rich Text Help Text Parser
  *
  *  This module contains a function loadHelpPage() and associated
  *  utilities to render (part of) an XML file into a RichDocument. 
  *  This is used for displaying help files.
  *
  *  Note that we assume valid, well-formed XML.
  *
  *  \todo When given an input such as
  *     <b>foo</b> <tt>bar</tt>
  *  (with arbitrary space between the tags) it will not render a
  *  space between "foo" and "bar".
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/richhelp.h"
#include "ui/richtext.h"
#include "util/xmlread.h"
#include "ui/richdocument.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/unicodechars.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "ui/skin.h"

namespace {
    /** XML to RichText parser. This class contains status information
        and functions for parsing RichText into XML. In particular, it
        saves a token type to implement a 1-lookahead parser. */
    class RichHelpRenderer {
        RichDocument& doc;
        XmlReader& rdr;
        XmlReader::Token current;

        void next();
        bool isOpen(const char* what);
        void flushLine(RichText& line);
        void parseBlock(int listLevel);
        void parseBulletList(int listLevel);
        void parseCountedList(int listLevel);
        void parseKeyList(int listLevel);
        void parseDefinitionList(int listLevel);
        void parseListItem(int listLevel);
        RichText parseText();
        RichText parseTextItem();
        void parseUnknown();

     public:
        RichHelpRenderer(RichDocument& doc, XmlReader& rdr);

        void parseDocument();
    };
}

/** Append string to RichText. This collapses multiple spaces into
    one, and strips newlines.
    \param out [in/out] RichText to append to
    \param haveSpace [in/out] Space status tracking
    \param in [in] Text to append */
static void
appendText(RichText& out, bool& haveSpace, const string_t& in)
{
    static const char SPACE[] = " \r\n\t";
    string_t::size_type a = 0;
    while (a < in.size()) {
        if (haveSpace) {
            /* out ends with a space. Skip all space in input. */
            a = in.find_first_not_of(SPACE, a);
            haveSpace = false;
        } else {
            /* out does not end with a space, so look for spaces in /in/. */
            string_t::size_type e = in.find_first_of(SPACE, a);
            if (e == in.npos) {
                out.append(string_t(in, a));
            } else {
                out.append(string_t(in, a, e-a));
                out.append(" ");
                haveSpace = true;
            }
            a = e;
        }
    }
}

/** Parse a color name. */
static UIColor::Color
parseColorName(const string_t& name)
{
    static const struct {
        const char* name;
        UIColor::Color color;
    } colors[] = {
        { "static", UIColor::tc_Static },
        { "green",  UIColor::tc_Green },
        { "yellow", UIColor::tc_Yellow },
        { "red",    UIColor::tc_Red },
        { "white",  UIColor::tc_White },
        { "blue",   UIColor::tc_Blue },
        { "dim",    UIColor::tc_Faded },
    };
    for (std::size_t i = 0; i < countof(colors); ++i) {
        if (name == colors[i].name) {
            return colors[i].color;
        }
    }
    return UIColor::tc_Static;
}

/**************************** RichHelpRenderer ***************************/

RichHelpRenderer::RichHelpRenderer(RichDocument& doc, XmlReader& rdr)
    : doc(doc),
      rdr(rdr),
      current(rdr.Eof)
{ }

/** Advance to next tag. */
inline void
RichHelpRenderer::next()
{
    current = rdr.read();
}

/** Check for opening tag. If that tag is found, skips it from the
    token stream.
    \param what Tag to check for (case sensitive!) */
bool
RichHelpRenderer::isOpen(const char* what)
{
    if (current == rdr.TagStart && rdr.getTag() == what) {
        next();
        return true;
    } else {
        return false;
    }
}

/** Output line to document if present. Clears it afterwards. */
void
RichHelpRenderer::flushLine(RichText& line)
{
    if (line.size()) {
        doc.add(line);
        doc.addNewline();
        line.clear();
    }
}

/** Parse a block-level element.
    - h1/h2/h3 headings
    - p paragraphs
    - ul/ol/kl/dl lists

    Upon call, we're looking at the opening tag.
    Upon exit, we're looking at the next element after the block element.

    \param listLevel list nesting level */
void
RichHelpRenderer::parseBlock(int listLevel)
{
    if (isOpen("h1")) {
        doc.add(parseText().withAttribute(RichTextStyleAttribute::Big).withAttribute(RichTextStyleAttribute::Bold));
        doc.addParagraph();
        next();
    } else if (isOpen("h2")) {
        doc.add(parseText().withAttribute(RichTextStyleAttribute::Bold));
        doc.addParagraph();
        next();
    } else if (isOpen("h3")) {
        doc.add(parseText().withAttribute(RichTextStyleAttribute::Underline));
        doc.addParagraph();
        next();
    } else if (isOpen("p")) {
        doc.add(parseText());
        doc.addParagraph();
        next();
    } else if (isOpen("ul")) {
        parseBulletList(listLevel);
        next();
    } else if (isOpen("ol")) {
        parseCountedList(listLevel);
        next();
    } else if (isOpen("dl")) {
        parseDefinitionList(listLevel);
        next();
    } else if (isOpen("kl")) {
        parseKeyList(listLevel);
        next();
    } else {
        parseUnknown();
    }
}

/** Parse bullet list.
    Upon call, we're looking at the first token of the content.
    Upon exit, we're looking at the closing tag. */
void
RichHelpRenderer::parseBulletList(int listLevel)
{
    static const char*const bullets[] = {
        UTF_SQUARE_BULLET,
        UTF_BULLET,
        "+",
        "-",
        ".",
    };

    const int em = font_ems[FONT_NORMAL] * 3/2;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("li")) {
            /* List item */
            int margin = doc.getLeftMargin();
            doc.addAt(margin, bullets[listLevel % countof(bullets)]);
            doc.add(" "); /* needed to make sure setLeftMargin does not indent the bullet! */
            doc.setLeftMargin(margin + em);
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            if (listLevel == 0) {
                doc.addNewline();
            }
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
}

/** Parse counted list. */
void
RichHelpRenderer::parseCountedList(int listLevel)
{
    const int em = font_ems[FONT_NORMAL] * 3/2;
    int counter = 0;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("li")) {
            /* List item */
            ++counter;
            int margin = doc.getLeftMargin();
            doc.addAt(margin, numToString(counter));
            doc.add(". ");
            doc.setLeftMargin(margin + em);
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            if (listLevel == 0) {
                doc.addNewline();
            }
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
}

/** Parse definition list. */
void
RichHelpRenderer::parseDefinitionList(int listLevel)
{
    const int em = font_ems[FONT_NORMAL] * 3/2;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("di")) {
            /* List item */
            string_t name;
            while (current == rdr.TagAttribute) {
                if (rdr.getName() == "term") {
                    name = rdr.getValue();
                }
                next();
            }
            int margin = doc.getLeftMargin();
            doc.addAt(margin, RichText(name).withAttribute(RichTextStyleAttribute::Bold));
            doc.add(" ");
            doc.setLeftMargin(margin + em);
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            if (listLevel == 0) {
                doc.addNewline();
            }
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
}

/** Parse key list. */
void
RichHelpRenderer::parseKeyList(int listLevel)
{
    const int em = font_ems[FONT_NORMAL] * 5;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("ki")) {
            /* List item */
            string_t name;
            while (current == rdr.TagAttribute) {
                if (rdr.getName() == "key") {
                    name = rdr.getValue();
                }
                next();
            }
            int margin = doc.getLeftMargin();
            doc.addAt(margin, RichText(name).withAttribute(RichTextStyleAttribute::Fixed));
            doc.tabTo(margin + em);
            doc.setLeftMargin(margin + em);
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
    if (listLevel == 0) {
        doc.addNewline();
    }
}

/** Parse list item.
    Upon call, we're looking at the first token of the content.
    Upon exit, we're looking at the closing tag of the list item. */
void
RichHelpRenderer::parseListItem(int listLevel)
{
    RichText line;
    bool haveSpace = true;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (current == rdr.Text) {
            /* Text */
            appendText(line, haveSpace, rdr.getValue());
            next();
        } else if (isOpen("ul")) {
            /* Sublist */
            flushLine(line);
            parseBulletList(listLevel);
            next();
            haveSpace = true;
        } else if (isOpen("ol")) {
            /* Sublist */
            flushLine(line);
            parseCountedList(listLevel);
            next();
            haveSpace = true;
        } else if (isOpen("dl")) {
            /* Sublist */
            flushLine(line);
            parseDefinitionList(listLevel);
            next();
            haveSpace = true;
        } else if (isOpen("kl")) {
            /* Sublist */
            flushLine(line);
            parseKeyList(listLevel);
            next();
            haveSpace = true;
        } else {
            /* Formatted text */
            line.append(parseTextItem());
            next();
            haveSpace = false;
        }
    }

    if (line.size()) {
        doc.add(line);
        doc.addNewline();
    }
}

/** Parse text element. */
RichText
RichHelpRenderer::parseText()
{
    /* Text consists of a, b, em, tt, key, font, where a and font have parameters. */
    bool haveSpace = true;
    RichText result;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (current == rdr.TagAttribute) {
            /* Spurious attribute from surrounding tag */
            next();
        } else if (current == rdr.Text) {
            /* Text */
            appendText(result, haveSpace, rdr.getValue());
            next();
        } else {
            result.append(parseTextItem());
            next();
            haveSpace = false;
        }
    }
    return result;
}

RichText
RichHelpRenderer::parseTextItem()
{
    if (isOpen("a")) {
        string_t target;
        while (current == rdr.TagAttribute) {
            if (rdr.getName() == "href") {
                target = rdr.getValue();
            }
            next();
        }
        return parseText().withAttribute(new RichTextLinkAttribute(target));
    } else if (isOpen("b")) {
        return parseText().withAttribute(RichTextStyleAttribute::Bold);
    } else if (isOpen("em")) {
        return parseText().withAttribute(RichTextStyleAttribute::Italic);
    } else if (isOpen("tt")) {
        return parseText().withAttribute(RichTextStyleAttribute::Fixed);
    } else if (isOpen("kbd") || isOpen("key")) {
        // FIXME: rendering not implemented yet
        return parseText().withAttribute(RichTextStyleAttribute::Fixed);
    } else if (isOpen("font")) {
        UIColor::Color c = UIColor::tc_Static;
        while (current == rdr.TagAttribute) {
            if (rdr.getName() == "color") {
                c = parseColorName(rdr.getValue());
            }
            next();
        }
        return parseText().withAttribute(new RichTextColorAttribute(c));
    } else {
        /* Skip it */
        return RichText();
    }
}

void
RichHelpRenderer::parseUnknown()
{
    /* Parse some unknown tag. When we're here, 'current' is the opening tag. */
    int nesting = 0;
    do {
        if (current == rdr.TagStart) {
            ++nesting;
        }
        if (current == rdr.TagEnd) {
            --nesting;
        }
        next();
    } while (current != rdr.Eof && nesting != 0);
}

void
RichHelpRenderer::parseDocument()
{
    /* A document is a list of block elements.
       It might be intermixed with
       - remaining attributes from the opening tag
       - spurious text elements (usually blank) */
    current = rdr.read();
    while (current != rdr.TagEnd && current != rdr.Eof && current != rdr.Error) {
        if (current == rdr.TagStart) {
            parseBlock(0);
        } else {
            next();
        }
    }
}


/****************************** Entry Point ******************************/

/** Load help page into a rich document.
    \param doc [out] The document
    \param rdr [in] XML Reader
    \param name [in] Name of help page */
void
loadHelpPage(RichDocument& doc, XmlReader& rdr, const string_t name)
{
    /* Locate help page */
    bool atPageStart = false;
    enum { Searching, Found, End } state = Searching;
    while (state == Searching) {
        switch (rdr.read()) {
         case XmlReader::TagStart:
            atPageStart = (rdr.getTag() == "page");
            break;
         case XmlReader::TagAttribute:
            if (atPageStart && rdr.getName() == "id" && rdr.getValue() == name) {
                state = Found;
            }
            break;
         case XmlReader::Eof:
            state = End;
            break;
         default:
            break;
        }
    }

    /* Found? */
    if (state != Found) {
        /* Not found. Render error page. */
        doc.add(RichText(_("Error")).withAttribute(RichTextStyleAttribute::Big).withAttribute(RichTextStyleAttribute::Bold));
        doc.addParagraph();
        doc.add(format(_("The help page \"%s\" could not be found within the help file."), name));
        doc.finish();
    } else {
        /* Found. Read and render it. */
        RichHelpRenderer(doc, rdr).parseDocument();
        doc.finish();
    }
}
