/**
  *  \file util/conf-file.cc
  *  \brief Configuration File Parsing
  *
  *  This implements classes to use a ConfigParser to fill a Config
  *  object.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/confpars.h"
#include "util/misc.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "cpluslib/string.h"
#include "cpluslib/smartptr.h"
#include "io/streams.h"
#include "util/conf.h"
#include "util/except.h"
#include "arch/platform.h"

/**************************** ConfigFileParser ***************************/

/** \class ConfigFileParser
    \brief Configuration File Parser

    This uses ConfigParser to fill a Config object. Errors in the
    configuration file will be logged as warnings on the console.  */

/** Constructor.
    \param the_config      The configuration object to fill
    \param ignore_unknown  If true, unknown config values will be silently ignored.
                           Otherwise, they will be logged on the console.
    \param src             Source flag for options */

ConfigFileParser::ConfigFileParser(Config& the_config, bool ignore_unknown, ConfigOption::Source src)
    : the_config(the_config),
      ignore_unknown(ignore_unknown),
      src(src)
{ }

void
ConfigFileParser::assign(int line_id, string_t name, string_t value, string_t /*line*/)
{
    try {
        ConfigOption* o = the_config.getConfigOption(name);
        if (!o) {
            if (!ignore_unknown)
                error(line_id, format(_("Unknown parameter '%s'"), name));
        } else {
            o->set(value);
            o->setSource(src);
        }
    }
    catch(std::domain_error& e) {
        error(line_id, e.what());
    }
    catch(std::runtime_error& e) {
        error(line_id, e.what());
    }
}

void
ConfigFileParser::error(int line_id, string_t msg)
{
    console.write(LOG_WARN, format(_("%s:%d: %s -- line has been ignored"), getFileName(), line_id, msg));
}

/*************************** ConfigFileUpdater ***************************/

/** \class ConfigFileUpdater
    \brief Configuration File Updater

    This uses ConfigParser to parse a configuration file, while at the
    same time generating an updated version of the configuration file
    in another stream. Comments and option order will be preserved.

    You must (optionally) call parseFile() etc. to process an existing
    file, followed by finish() to write out all config values not yet
    contained in the original file. */

/** Constructor.
    \param config Configuration object
    \param out Stream to write new file to */
ConfigFileUpdater::ConfigFileUpdater(Config& the_config, Stream& out)
    : the_config(the_config), out(out)
{ }

void
ConfigFileUpdater::assign(int /*line_id*/, string_t name, string_t /*value*/, string_t line)
{
    ConfigOption* o = the_config.getConfigOption(name);
    if (o) {
        // FIXME: can we take over the indentation?
        // FIXME2: only write modified options. This needs source tracking beyond set/unset
        line = o->getName();
        line += " = ";
        line += o->toString();
        written_options.insert(o);
    }

    line += Platform::LINE_SEPARATOR;
    out.writeT(line.data(), line.size());
}

void
ConfigFileUpdater::error(int /*line_id*/, string_t /*msg*/)
{
    // Ignore all errors. An error occurs when a line within our section has no "key=" at the beginning.
}

void
ConfigFileUpdater::unparsed(int /*line_id*/, string_t line)
{
    line += Platform::LINE_SEPARATOR;
    out.writeT(line.data(), line.size());
}

/** Finish new config file. If not all options have been written out yet,
    they will be written now. */
void
ConfigFileUpdater::finish()
{
    bool had_section_delimiter = false;

    for (Config::iterator i = the_config.begin(); i != the_config.end(); ++i) {
        if (written_options.find(&*i) == written_options.end()) {
            // Write section header if required
            if (!isInSection() && !had_section_delimiter) {
                string_t line = Platform::LINE_SEPARATOR;
                line += "% ";
                line += getSectionName();
                line += Platform::LINE_SEPARATOR;
                line += Platform::LINE_SEPARATOR;
                out.writeT(line.data(), line.size());
                had_section_delimiter = true;
            }

            // Write option value
            string_t line = i->getName();
            line += " = ";
            line += i->toString();
            line += Platform::LINE_SEPARATOR;
            out.writeT(line.data(), line.size());
        }
    }
}
